# pylint:disable=unused-argument,no-self-use
from __future__ import annotations
from abc import abstractmethod
from typing import Any, Generic, TypeVar, cast
from collections.abc import Callable

from . import Block
from .statement import (
    CAS,
    Statement,
    SideEffectStatement,
    ConditionalJump,
    Assignment,
    Store,
    Return,
    Jump,
    DirtyStatement,
    WeakAssignment,
)
from .expression import (
    Call,
    Load,
    Expression,
    BinaryOp,
    UnaryOp,
    Convert,
    ITE,
    DirtyExpression,
    VEXCCallExpression,
    Tmp,
    Register,
    Const,
    Reinterpret,
    MultiStatementExpression,
    VirtualVariable,
    Phi,
    Atom,
    Extract,
    Insert,
)

ExprType = TypeVar("ExprType")
StmtType = TypeVar("StmtType")
BlockType = TypeVar("BlockType")


class AILBlockWalker(Generic[ExprType, StmtType, BlockType]):
    """
    Walks all statements and expressions of an AIL node and construct arbitrary values based on them.
    """

    def __init__(self, stmt_handlers=None, expr_handlers=None):
        _default_stmt_handlers: dict[type, Callable[[int, Any, Block | None], StmtType]] = {
            Assignment: self._handle_Assignment,
            WeakAssignment: self._handle_WeakAssignment,
            CAS: self._handle_CAS,
            SideEffectStatement: self._handle_SideEffectStatement,
            Store: self._handle_Store,
            ConditionalJump: self._handle_ConditionalJump,
            Jump: self._handle_Jump,
            Return: self._handle_Return,
            DirtyStatement: self._handle_DirtyStatement,
        }

        _default_expr_handlers: dict[type, Callable[[int, Any, int, Statement | None, Block | None], ExprType]] = {
            Call: self._handle_CallExpr,
            Load: self._handle_Load,
            BinaryOp: self._handle_BinaryOp,
            UnaryOp: self._handle_UnaryOp,
            Convert: self._handle_Convert,
            ITE: self._handle_ITE,
            DirtyExpression: self._handle_DirtyExpression,
            VEXCCallExpression: self._handle_VEXCCallExpression,
            Tmp: self._handle_Tmp,
            Register: self._handle_Register,
            Reinterpret: self._handle_Reinterpret,
            Const: self._handle_Const,
            MultiStatementExpression: self._handle_MultiStatementExpression,
            VirtualVariable: self._handle_VirtualVariable,
            Phi: self._handle_Phi,
            Extract: self._handle_Extract,
            Insert: self._handle_Insert,
        }

        self.stmt_handlers: dict[type, Callable[[int, Any, Block | None], StmtType]] = (
            stmt_handlers or _default_stmt_handlers
        )
        self.expr_handlers: dict[type, Callable[[int, Any, int, Statement | None, Block | None], ExprType]] = (
            expr_handlers or _default_expr_handlers
        )

    def walk(self, block: Block) -> BlockType:
        i = 0
        results = []
        while i < len(block.statements):
            stmt = block.statements[i]
            results.append(self._handle_stmt(i, stmt, block))
            i += 1
        return self._handle_block_end(results, block)

    @abstractmethod
    def _handle_block_end(self, stmt_results: list[StmtType], block: Block) -> BlockType:
        raise NotImplementedError

    def walk_statement(self, stmt: Statement, block: Block | None = None, stmt_idx: int = 0) -> StmtType:
        return self._handle_stmt(stmt_idx, stmt, block)

    def walk_expression(
        self,
        expr: Expression,
        stmt_idx: int | None = None,
        stmt: Statement | None = None,
        block: Block | None = None,
    ) -> ExprType:
        return self._handle_expr(0, expr, stmt_idx or 0, stmt, block)

    def _handle_stmt(self, stmt_idx: int, stmt: Statement, block: Block | None) -> StmtType:
        handler = self.stmt_handlers.get(type(stmt), self._stmt_top)
        return handler(stmt_idx, stmt, block)

    def _handle_expr(
        self, expr_idx: int, expr: Expression, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> ExprType:
        handler = self.expr_handlers.get(type(expr), self._top)
        return handler(expr_idx, expr, stmt_idx, stmt, block)

    @abstractmethod
    def _top(
        self, expr_idx: int, expr: Expression, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> ExprType:
        raise NotImplementedError

    @abstractmethod
    def _stmt_top(self, stmt_idx: int, stmt: Statement, block: Block | None) -> StmtType:
        raise NotImplementedError

    #
    # Default handlers
    #

    def _handle_Assignment(self, stmt_idx: int, stmt: Assignment, block: Block | None) -> StmtType:
        self._handle_expr(0, stmt.dst, stmt_idx, stmt, block)
        self._handle_expr(1, stmt.src, stmt_idx, stmt, block)
        return self._stmt_top(stmt_idx, stmt, block)

    def _handle_WeakAssignment(self, stmt_idx: int, stmt: WeakAssignment, block: Block | None) -> StmtType:
        self._handle_expr(0, stmt.dst, stmt_idx, stmt, block)
        self._handle_expr(1, stmt.src, stmt_idx, stmt, block)
        return self._stmt_top(stmt_idx, stmt, block)

    def _handle_CAS(self, stmt_idx: int, stmt: CAS, block: Block | None) -> StmtType:
        self._handle_expr(0, stmt.addr, stmt_idx, stmt, block)
        self._handle_expr(1, stmt.data_lo, stmt_idx, stmt, block)
        if stmt.data_hi is not None:
            self._handle_expr(2, stmt.data_hi, stmt_idx, stmt, block)
        self._handle_expr(3, stmt.expd_lo, stmt_idx, stmt, block)
        if stmt.expd_hi is not None:
            self._handle_expr(4, stmt.expd_hi, stmt_idx, stmt, block)
        self._handle_expr(5, stmt.old_lo, stmt_idx, stmt, block)
        if stmt.old_hi is not None:
            self._handle_expr(6, stmt.old_hi, stmt_idx, stmt, block)
        return self._stmt_top(stmt_idx, stmt, block)

    def _handle_SideEffectStatement(self, stmt_idx: int, stmt: SideEffectStatement, block: Block | None) -> StmtType:
        self._handle_expr(0, stmt.expr, stmt_idx, stmt, block)
        return self._stmt_top(stmt_idx, stmt, block)

    def _handle_Store(self, stmt_idx: int, stmt: Store, block: Block | None) -> StmtType:
        self._handle_expr(0, stmt.addr, stmt_idx, stmt, block)
        self._handle_expr(1, stmt.data, stmt_idx, stmt, block)
        if stmt.guard is not None:
            self._handle_expr(2, stmt.guard, stmt_idx, stmt, block)
        return self._stmt_top(stmt_idx, stmt, block)

    def _handle_Jump(self, stmt_idx: int, stmt: Jump, block: Block | None) -> StmtType:
        self._handle_expr(0, stmt.target, stmt_idx, stmt, block)
        return self._stmt_top(stmt_idx, stmt, block)

    def _handle_ConditionalJump(self, stmt_idx: int, stmt: ConditionalJump, block: Block | None) -> StmtType:
        self._handle_expr(0, stmt.condition, stmt_idx, stmt, block)
        if stmt.true_target is not None:
            self._handle_expr(1, stmt.true_target, stmt_idx, stmt, block)
        if stmt.false_target is not None:
            self._handle_expr(2, stmt.false_target, stmt_idx, stmt, block)
        return self._stmt_top(stmt_idx, stmt, block)

    def _handle_Return(self, stmt_idx: int, stmt: Return, block: Block | None) -> StmtType:
        if stmt.ret_exprs:
            for i, ret_expr in enumerate(stmt.ret_exprs):
                self._handle_expr(i, ret_expr, stmt_idx, stmt, block)
        return self._stmt_top(stmt_idx, stmt, block)

    def _handle_DirtyStatement(self, stmt_idx: int, stmt: DirtyStatement, block: Block | None) -> StmtType:
        self._handle_expr(0, stmt.dirty, stmt_idx, stmt, block)
        return self._stmt_top(stmt_idx, stmt, block)

    def _handle_Load(
        self, expr_idx: int, expr: Load, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> ExprType:
        self._handle_expr(0, expr.addr, stmt_idx, stmt, block)
        return self._top(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_CallExpr(
        self, expr_idx: int, expr: Call, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> ExprType:
        if not isinstance(expr.target, str):
            self._handle_expr(-1, expr.target, stmt_idx, stmt, block)
        if expr.args:
            for i, arg in enumerate(expr.args):
                self._handle_expr(i, arg, stmt_idx, stmt, block)
        return self._top(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_BinaryOp(
        self, expr_idx: int, expr: BinaryOp, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> ExprType:
        self._handle_expr(0, expr.operands[0], stmt_idx, stmt, block)
        self._handle_expr(1, expr.operands[1], stmt_idx, stmt, block)
        return self._top(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_UnaryOp(
        self, expr_idx: int, expr: UnaryOp, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> ExprType:
        self._handle_expr(0, expr.operand, stmt_idx, stmt, block)
        return self._top(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_Convert(
        self, expr_idx: int, expr: Convert, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> ExprType:
        self._handle_expr(expr_idx, expr.operand, stmt_idx, stmt, block)
        return self._top(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_Reinterpret(
        self, expr_idx: int, expr: Reinterpret, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> ExprType:
        self._handle_expr(expr_idx, expr.operand, stmt_idx, stmt, block)
        return self._top(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_ITE(
        self, expr_idx: int, expr: ITE, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> ExprType:
        self._handle_expr(0, expr.cond, stmt_idx, stmt, block)
        self._handle_expr(1, expr.iftrue, stmt_idx, stmt, block)
        self._handle_expr(2, expr.iffalse, stmt_idx, stmt, block)
        return self._top(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_Tmp(
        self, expr_idx: int, expr: Tmp, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> ExprType:
        return self._top(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_Register(
        self, expr_idx: int, expr: Register, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> ExprType:
        return self._top(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_Const(
        self, expr_idx: int, expr: Const, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> ExprType:
        return self._top(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_VirtualVariable(
        self, expr_idx: int, expr: VirtualVariable, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> ExprType:
        return self._top(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_Phi(
        self, expr_idx: int, expr: Phi, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> ExprType:
        for idx, (_, vvar) in enumerate(expr.src_and_vvars):
            if vvar is not None:
                self._handle_expr(idx, vvar, stmt_idx, stmt, block)
        return self._top(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_MultiStatementExpression(
        self, expr_idx, expr: MultiStatementExpression, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> ExprType:
        for idx, stmt_ in enumerate(expr.stmts):
            self._handle_stmt(idx, stmt_, None)
        self._handle_expr(0, expr.expr, stmt_idx, stmt, block)
        return self._top(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_DirtyExpression(
        self, expr_idx: int, expr: DirtyExpression, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> ExprType:
        for idx, operand in enumerate(expr.operands):
            self._handle_expr(idx, operand, stmt_idx, stmt, block)
        if expr.guard is not None:
            self._handle_expr(len(expr.operands) + 1, expr.guard, stmt_idx, stmt, block)
        return self._top(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_VEXCCallExpression(
        self, expr_idx: int, expr: VEXCCallExpression, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> ExprType:
        for idx, operand in enumerate(expr.operands):
            self._handle_expr(idx, operand, stmt_idx, stmt, block)
        return self._top(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_Extract(
        self, expr_idx: int, expr: Extract, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> ExprType:
        self._handle_expr(0, expr.base, stmt_idx, stmt, block)
        self._handle_expr(1, expr.offset, stmt_idx, stmt, block)
        return self._top(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_Insert(
        self, expr_idx: int, expr: Insert, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> ExprType:
        self._handle_expr(0, expr.base, stmt_idx, stmt, block)
        self._handle_expr(1, expr.offset, stmt_idx, stmt, block)
        self._handle_expr(2, expr.value, stmt_idx, stmt, block)
        return self._top(expr_idx, expr, stmt_idx, stmt, block)


class AILBlockViewer(AILBlockWalker[None, None, None]):
    """
    Walks all statements and expressions of an AIL node and do nothing.
    """

    def _top(self, expr_idx: int, expr: Expression, stmt_idx: int, stmt: Statement | None, block: Block | None):
        return None

    def _stmt_top(self, stmt_idx: int, stmt: Statement, block: Block | None):
        return None

    def _handle_block_end(self, stmt_results: list[StmtType], block: Block):
        return None

    # Duplicate all handlers for performance...

    def _handle_Assignment(self, stmt_idx: int, stmt: Assignment, block: Block | None):
        self._handle_expr(0, stmt.dst, stmt_idx, stmt, block)
        self._handle_expr(1, stmt.src, stmt_idx, stmt, block)

    def _handle_WeakAssignment(self, stmt_idx: int, stmt: WeakAssignment, block: Block | None):
        self._handle_expr(0, stmt.dst, stmt_idx, stmt, block)
        self._handle_expr(1, stmt.src, stmt_idx, stmt, block)

    def _handle_CAS(self, stmt_idx: int, stmt: CAS, block: Block | None):
        self._handle_expr(0, stmt.addr, stmt_idx, stmt, block)
        self._handle_expr(1, stmt.data_lo, stmt_idx, stmt, block)
        if stmt.data_hi is not None:
            self._handle_expr(2, stmt.data_hi, stmt_idx, stmt, block)
        self._handle_expr(3, stmt.expd_lo, stmt_idx, stmt, block)
        if stmt.expd_hi is not None:
            self._handle_expr(4, stmt.expd_hi, stmt_idx, stmt, block)
        self._handle_expr(5, stmt.old_lo, stmt_idx, stmt, block)
        if stmt.old_hi is not None:
            self._handle_expr(6, stmt.old_hi, stmt_idx, stmt, block)

    def _handle_SideEffectStatement(self, stmt_idx: int, stmt: SideEffectStatement, block: Block | None):
        self._handle_expr(0, stmt.expr, stmt_idx, stmt, block)

    def _handle_Store(self, stmt_idx: int, stmt: Store, block: Block | None):
        self._handle_expr(0, stmt.addr, stmt_idx, stmt, block)
        self._handle_expr(1, stmt.data, stmt_idx, stmt, block)
        if stmt.guard is not None:
            self._handle_expr(2, stmt.guard, stmt_idx, stmt, block)

    def _handle_Jump(self, stmt_idx: int, stmt: Jump, block: Block | None):
        self._handle_expr(0, stmt.target, stmt_idx, stmt, block)

    def _handle_ConditionalJump(self, stmt_idx: int, stmt: ConditionalJump, block: Block | None):
        self._handle_expr(0, stmt.condition, stmt_idx, stmt, block)
        if stmt.true_target is not None:
            self._handle_expr(1, stmt.true_target, stmt_idx, stmt, block)
        if stmt.false_target is not None:
            self._handle_expr(2, stmt.false_target, stmt_idx, stmt, block)

    def _handle_Return(self, stmt_idx: int, stmt: Return, block: Block | None):
        if stmt.ret_exprs:
            for i, ret_expr in enumerate(stmt.ret_exprs):
                self._handle_expr(i, ret_expr, stmt_idx, stmt, block)

    def _handle_DirtyStatement(self, stmt_idx: int, stmt: DirtyStatement, block: Block | None):
        self._handle_expr(0, stmt.dirty, stmt_idx, stmt, block)

    def _handle_Load(self, expr_idx: int, expr: Load, stmt_idx: int, stmt: Statement | None, block: Block | None):
        self._handle_expr(0, expr.addr, stmt_idx, stmt, block)

    def _handle_CallExpr(self, expr_idx: int, expr: Call, stmt_idx: int, stmt: Statement | None, block: Block | None):
        if not isinstance(expr.target, str):
            self._handle_expr(-1, expr.target, stmt_idx, stmt, block)
        if expr.args:
            for i, arg in enumerate(expr.args):
                self._handle_expr(i, arg, stmt_idx, stmt, block)

    def _handle_BinaryOp(
        self, expr_idx: int, expr: BinaryOp, stmt_idx: int, stmt: Statement | None, block: Block | None
    ):
        self._handle_expr(0, expr.operands[0], stmt_idx, stmt, block)
        self._handle_expr(1, expr.operands[1], stmt_idx, stmt, block)

    def _handle_UnaryOp(self, expr_idx: int, expr: UnaryOp, stmt_idx: int, stmt: Statement | None, block: Block | None):
        self._handle_expr(0, expr.operand, stmt_idx, stmt, block)

    def _handle_Convert(self, expr_idx: int, expr: Convert, stmt_idx: int, stmt: Statement | None, block: Block | None):
        self._handle_expr(expr_idx, expr.operand, stmt_idx, stmt, block)

    def _handle_Reinterpret(
        self, expr_idx: int, expr: Reinterpret, stmt_idx: int, stmt: Statement | None, block: Block | None
    ):
        self._handle_expr(expr_idx, expr.operand, stmt_idx, stmt, block)

    def _handle_ITE(self, expr_idx: int, expr: ITE, stmt_idx: int, stmt: Statement | None, block: Block | None):
        self._handle_expr(0, expr.cond, stmt_idx, stmt, block)
        self._handle_expr(1, expr.iftrue, stmt_idx, stmt, block)
        self._handle_expr(2, expr.iffalse, stmt_idx, stmt, block)

    def _handle_Tmp(self, expr_idx: int, expr: Tmp, stmt_idx: int, stmt: Statement | None, block: Block | None):
        return None

    def _handle_Register(
        self, expr_idx: int, expr: Register, stmt_idx: int, stmt: Statement | None, block: Block | None
    ):
        return None

    def _handle_Const(self, expr_idx: int, expr: Const, stmt_idx: int, stmt: Statement | None, block: Block | None):
        return None

    def _handle_VirtualVariable(
        self, expr_idx: int, expr: VirtualVariable, stmt_idx: int, stmt: Statement | None, block: Block | None
    ):
        return None

    def _handle_Phi(self, expr_idx: int, expr: Phi, stmt_idx: int, stmt: Statement | None, block: Block | None):
        for idx, (_, vvar) in enumerate(expr.src_and_vvars):
            if vvar is not None:
                self._handle_expr(idx, vvar, stmt_idx, stmt, block)

    def _handle_MultiStatementExpression(
        self, expr_idx, expr: MultiStatementExpression, stmt_idx: int, stmt: Statement | None, block: Block | None
    ):
        for idx, stmt_ in enumerate(expr.stmts):
            self._handle_stmt(idx, stmt_, None)
        self._handle_expr(0, expr.expr, stmt_idx, stmt, block)

    def _handle_DirtyExpression(
        self, expr_idx: int, expr: DirtyExpression, stmt_idx: int, stmt: Statement | None, block: Block | None
    ):
        for idx, operand in enumerate(expr.operands):
            self._handle_expr(idx, operand, stmt_idx, stmt, block)
        if expr.guard is not None:
            self._handle_expr(len(expr.operands) + 1, expr.guard, stmt_idx, stmt, block)

    def _handle_VEXCCallExpression(
        self, expr_idx: int, expr: VEXCCallExpression, stmt_idx: int, stmt: Statement | None, block: Block | None
    ):
        for idx, operand in enumerate(expr.operands):
            self._handle_expr(idx, operand, stmt_idx, stmt, block)

    def _handle_Extract(self, expr_idx: int, expr: Extract, stmt_idx: int, stmt: Statement | None, block: Block | None):
        self._handle_expr(0, expr.base, stmt_idx, stmt, block)
        self._handle_expr(1, expr.offset, stmt_idx, stmt, block)

    def _handle_Insert(self, expr_idx: int, expr: Insert, stmt_idx: int, stmt: Statement | None, block: Block | None):
        self._handle_expr(0, expr.base, stmt_idx, stmt, block)
        self._handle_expr(1, expr.offset, stmt_idx, stmt, block)
        self._handle_expr(2, expr.value, stmt_idx, stmt, block)


class AILBlockRewriter(AILBlockWalker[Expression, Statement, Block]):
    """
    Walks all statements and expressions of an AIL node, and rebuilds expressions, statements, or blocks if needed.

    If you need a pure walker without rebuilding, use AILBlockViewer instead.

    :ivar update_block: True if the block should be updated in place, False if a new block should be created and
                        returned as the result of walk().
    :ivar replace_phi_stmt: True if you want _handle_Phi be called and vvars potentially replaced; False otherwise.
                            Default to False because in the most majority cases you do not want vvars in a Phi
                            variable be replaced.
    """

    def __init__(
        self, stmt_handlers=None, expr_handlers=None, update_block: bool = True, replace_phi_stmt: bool = False
    ):
        super().__init__(stmt_handlers=stmt_handlers, expr_handlers=expr_handlers)
        self._update_block = update_block
        self._replace_phi_stmt = replace_phi_stmt

    def _top(
        self, expr_idx: int, expr: Expression, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> Expression:
        return expr

    def _stmt_top(self, stmt_idx: int, stmt: Statement, block: Block | None) -> Statement:
        return stmt

    def _handle_block_end(self, stmt_results: list[Statement], block: Block) -> Block:
        if all(new is None or new is old for new, old in zip(stmt_results, block.statements)):
            return block
        statements = [new or old for new, old in zip(stmt_results, block.statements)]
        if not self._update_block:
            return block.copy(statements=statements)
        block.statements = statements
        return block

    #
    # Default handlers
    #

    def _handle_Assignment(self, stmt_idx: int, stmt: Assignment, block: Block | None) -> Statement:
        dst = self._handle_expr(0, stmt.dst, stmt_idx, stmt, block)
        assert isinstance(dst, Atom)
        changed = dst is not stmt.dst

        src = self._handle_expr(1, stmt.src, stmt_idx, stmt, block)
        changed |= src is not stmt.src

        if changed:
            return Assignment(stmt.idx, dst, src, **stmt.tags)
        return stmt

    def _handle_WeakAssignment(self, stmt_idx: int, stmt: WeakAssignment, block: Block | None) -> Statement:
        dst = self._handle_expr(0, stmt.dst, stmt_idx, stmt, block)
        assert isinstance(dst, Atom)
        changed = dst is not stmt.dst

        src = self._handle_expr(1, stmt.src, stmt_idx, stmt, block)
        changed |= src is not stmt.src

        if changed:
            return WeakAssignment(stmt.idx, dst, src, **stmt.tags)
        return stmt

    def _handle_CAS(self, stmt_idx: int, stmt: CAS, block: Block | None) -> Statement:
        addr = self._handle_expr(0, stmt.addr, stmt_idx, stmt, block)
        changed = addr is not stmt.addr

        data_lo = self._handle_expr(1, stmt.data_lo, stmt_idx, stmt, block)
        changed |= data_lo is not stmt.data_lo

        data_hi = None
        if stmt.data_hi is not None:
            data_hi = self._handle_expr(2, stmt.data_hi, stmt_idx, stmt, block)
            changed |= data_hi is not stmt.data_hi

        expd_lo = self._handle_expr(3, stmt.expd_lo, stmt_idx, stmt, block)
        changed |= expd_lo is not stmt.expd_lo

        expd_hi = None
        if stmt.expd_hi is not None:
            expd_hi = self._handle_expr(4, stmt.expd_hi, stmt_idx, stmt, block)
            changed |= expd_hi is not stmt.expd_hi

        old_lo = self._handle_expr(5, stmt.old_lo, stmt_idx, stmt, block)
        assert isinstance(old_lo, Atom)
        changed |= old_lo is not stmt.old_lo

        old_hi = None
        if stmt.old_hi is not None:
            old_hi = self._handle_expr(6, stmt.old_hi, stmt_idx, stmt, block)
            assert isinstance(old_hi, Atom)
            changed |= old_hi is not stmt.old_hi

        if changed:
            return CAS(
                stmt.idx,
                addr,
                data_lo,
                data_hi,
                expd_lo,
                expd_hi,
                old_lo,
                old_hi,
                stmt.endness,
                **stmt.tags,
            )
        return stmt

    def _handle_SideEffectStatement(self, stmt_idx: int, stmt: SideEffectStatement, block: Block | None) -> Statement:
        new_expr = self._handle_expr(0, stmt.expr, stmt_idx, stmt, block)
        changed = new_expr is not stmt.expr

        if changed:
            return SideEffectStatement(
                stmt.idx,
                new_expr,
                ret_expr=stmt.ret_expr,
                fp_ret_expr=stmt.fp_ret_expr,
                **stmt.tags,
            )
        return stmt

    def _handle_Store(self, stmt_idx: int, stmt: Store, block: Block | None) -> Statement:
        addr = self._handle_expr(0, stmt.addr, stmt_idx, stmt, block)
        changed = addr is not stmt.addr

        data = self._handle_expr(1, stmt.data, stmt_idx, stmt, block)
        changed |= data is not stmt.data

        guard = None if stmt.guard is None else self._handle_expr(2, stmt.guard, stmt_idx, stmt, block)
        changed |= guard is not stmt.guard

        if changed:
            return Store(
                stmt.idx,
                addr,
                data,
                stmt.size,
                stmt.endness,
                guard=guard,
                variable=stmt.variable,
                offset=stmt.offset,
                **stmt.tags,
            )
        return stmt

    def _handle_Jump(self, stmt_idx: int, stmt: Jump, block: Block | None) -> Statement:
        target = self._handle_expr(0, stmt.target, stmt_idx, stmt, block)
        changed = target is not stmt.target

        if changed:
            return Jump(
                stmt.idx,
                target,
                target_idx=stmt.target_idx,
                **stmt.tags,
            )
        return stmt

    def _handle_ConditionalJump(self, stmt_idx: int, stmt: ConditionalJump, block: Block | None) -> Statement:
        condition = self._handle_expr(0, stmt.condition, stmt_idx, stmt, block)
        changed = condition is not stmt.condition

        true_target = None
        if stmt.true_target is not None:
            true_target = self._handle_expr(1, stmt.true_target, stmt_idx, stmt, block)
            changed |= true_target is not stmt.true_target

        false_target = None
        if stmt.false_target is not None:
            false_target = self._handle_expr(2, stmt.false_target, stmt_idx, stmt, block)
            changed |= false_target is not stmt.false_target

        if changed:
            return ConditionalJump(
                stmt.idx,
                condition,
                true_target,
                false_target,
                true_target_idx=stmt.true_target_idx,
                false_target_idx=stmt.false_target_idx,
                **stmt.tags,
            )
        return stmt

    def _handle_Return(self, stmt_idx: int, stmt: Return, block: Block | None) -> Statement:
        if stmt.ret_exprs:
            new_ret_exprs = [
                self._handle_expr(idx, expr, stmt_idx, stmt, block) for idx, expr in enumerate(stmt.ret_exprs)
            ]
            changed = any(old is not new for new, old in zip(new_ret_exprs, stmt.ret_exprs))

            if changed:
                return Return(stmt.idx, new_ret_exprs, **stmt.tags)
        return stmt

    def _handle_DirtyStatement(self, stmt_idx: int, stmt: DirtyStatement, block: Block | None) -> Statement:
        dirty = self._handle_expr(0, stmt.dirty, stmt_idx, stmt, block)
        assert isinstance(dirty, DirtyExpression)
        changed = dirty is not stmt.dirty

        if changed:
            return DirtyStatement(stmt.idx, dirty, **stmt.tags)
        return stmt

    #
    # Expression handlers

    def _handle_expr(self, expr_idx: int, expr: Expression, stmt_idx: int, stmt: Statement | None, block: Block | None):
        # reach a fixed point
        while True:
            result = super()._handle_expr(expr_idx, expr, stmt_idx, stmt, block)
            if result is expr:
                break
            expr = result
        return expr

    def _handle_Load(
        self, expr_idx: int, expr: Load, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> Expression:
        addr = self._handle_expr(0, expr.addr, stmt_idx, stmt, block)
        changed = addr is not expr.addr

        if changed:
            new_expr = expr.copy()
            new_expr.addr = addr
            return new_expr
        return expr

    def _handle_CallExpr(
        self, expr_idx: int, expr: Call, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> Expression:
        changed = False

        if isinstance(expr.target, str):
            new_target = expr.target
        else:
            new_target = self._handle_expr(-1, expr.target, stmt_idx, stmt, block)
            changed |= new_target is not expr.target

        new_args = None
        if expr.args is not None:
            new_args = [self._handle_expr(idx, arg, stmt_idx, stmt, block) for idx, arg in enumerate(expr.args)]
            changed |= any(old is not new for new, old in zip(new_args, expr.args))

        if changed:
            expr = expr.copy()
            expr.target = new_target
            expr.args = new_args
            return expr
        return expr

    def _handle_BinaryOp(
        self, expr_idx: int, expr: BinaryOp, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> Expression:
        operand_0 = self._handle_expr(0, expr.operands[0], stmt_idx, stmt, block)
        changed = operand_0 is not expr.operands[0]

        operand_1 = self._handle_expr(1, expr.operands[1], stmt_idx, stmt, block)
        changed |= operand_1 is not expr.operands[1]

        if changed:
            new_expr = expr.copy()
            new_expr.operands = (operand_0, operand_1)
            new_expr.depth = max(operand_0.depth, operand_1.depth) + 1
            return new_expr
        return expr

    def _handle_UnaryOp(
        self, expr_idx: int, expr: UnaryOp, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> Expression:
        new_operand = self._handle_expr(0, expr.operand, stmt_idx, stmt, block)
        changed = new_operand is not expr.operand

        if changed:
            new_expr = expr.copy()
            new_expr.operand = new_operand
            return new_expr
        return expr

    def _handle_Convert(
        self, expr_idx: int, expr: Convert, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> Expression:
        new_operand = self._handle_expr(expr_idx, expr.operand, stmt_idx, stmt, block)
        changed = new_operand is not expr.operand

        if changed:
            return Convert(expr.idx, expr.from_bits, expr.to_bits, expr.is_signed, new_operand, **expr.tags)
        return expr

    def _handle_Reinterpret(
        self, expr_idx: int, expr: Reinterpret, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> Expression:
        new_operand = self._handle_expr(expr_idx, expr.operand, stmt_idx, stmt, block)
        changed = new_operand is not expr.operand

        if changed:
            return Reinterpret(
                expr.idx, expr.from_bits, expr.from_type, expr.to_bits, expr.to_type, new_operand, **expr.tags
            )
        return expr

    def _handle_ITE(
        self, expr_idx: int, expr: ITE, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> Expression:
        cond = self._handle_expr(0, expr.cond, stmt_idx, stmt, block)
        changed = cond is not expr.cond

        iftrue = self._handle_expr(1, expr.iftrue, stmt_idx, stmt, block)
        changed |= iftrue is not expr.iftrue

        iffalse = self._handle_expr(2, expr.iffalse, stmt_idx, stmt, block)
        changed |= iffalse is not expr.iffalse

        if changed:
            new_expr = expr.copy()
            new_expr.cond = cond
            new_expr.iftrue = iftrue
            new_expr.iffalse = iffalse
            return new_expr
        return expr

    def _handle_Phi(
        self, expr_idx: int, expr: Phi, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> Expression:
        if not self._replace_phi_stmt:
            # fallback to the read-only version
            super()._handle_Phi(expr_idx, expr, stmt_idx, stmt, block)
            return expr

        changed = False

        src_and_vvars = [
            (src, self._handle_expr(idx, vvar, stmt_idx, stmt, block) if vvar is not None else None)
            for idx, (src, vvar) in enumerate(expr.src_and_vvars)
        ]
        changed = any(new is not old for (_, new), (_, old) in zip(src_and_vvars, expr.src_and_vvars))

        if changed:
            assert all(vvar is None or isinstance(vvar, VirtualVariable) for _, vvar in src_and_vvars)
            return Phi(
                expr.idx,
                expr.bits,
                cast(list[tuple[tuple[int, int | None], VirtualVariable | None]], src_and_vvars),
                **expr.tags,
            )
        return expr

    def _handle_DirtyExpression(
        self, expr_idx: int, expr: DirtyExpression, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> Expression:
        changed = False
        new_operands = [self._handle_expr(0, operand, stmt_idx, stmt, block) for operand in expr.operands]
        changed = any(new is not old for new, old in zip(new_operands, expr.operands))

        new_guard = None
        if expr.guard is not None:
            new_guard = self._handle_expr(2, expr.guard, stmt_idx, stmt, block)
            changed |= new_guard is not expr.guard

        if changed:
            return DirtyExpression(
                expr.idx,
                expr.callee,
                new_operands,
                guard=new_guard,
                mfx=expr.mfx,
                maddr=expr.maddr,
                msize=expr.msize,
                bits=expr.bits,
                **expr.tags,
            )
        return expr

    def _handle_VEXCCallExpression(
        self, expr_idx: int, expr: VEXCCallExpression, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> Expression:
        new_operands = [
            self._handle_expr(idx, operand, stmt_idx, stmt, block) for idx, operand in enumerate(expr.operands)
        ]
        changed = any(new is not old for new, old in zip(new_operands, expr.operands))

        if changed:
            new_expr = expr.copy()
            new_expr.operands = tuple(new_operands)
            return new_expr
        return expr

    def _handle_MultiStatementExpression(
        self, expr_idx, expr: MultiStatementExpression, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> Expression:
        new_statements = [self._handle_stmt(idx, stmt_, None) for idx, stmt_ in enumerate(expr.stmts)]
        changed = any(new is not old for new, old in zip(new_statements, expr.stmts))

        new_expr = self._handle_expr(0, expr.expr, stmt_idx, stmt, block)
        changed |= new_expr is not expr.expr

        if changed:
            expr_ = expr.copy()
            expr_.expr = new_expr
            expr_.stmts = new_statements
            return expr_
        return expr

    def _handle_Extract(
        self, expr_idx: int, expr: Extract, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> Expression:
        new_base = self._handle_expr(0, expr.base, stmt_idx, stmt, block)
        new_offset = self._handle_expr(1, expr.offset, stmt_idx, stmt, block)

        if new_base is not expr.base or new_offset is not expr.offset:
            result = expr.copy()
            result.base = new_base
            result.offset = new_offset
            return result
        return expr

    def _handle_Insert(
        self, expr_idx: int, expr: Insert, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> Expression:
        new_base = self._handle_expr(0, expr.base, stmt_idx, stmt, block)
        new_offset = self._handle_expr(1, expr.offset, stmt_idx, stmt, block)
        new_value = self._handle_expr(2, expr.value, stmt_idx, stmt, block)

        if new_base is not expr.base or new_offset is not expr.offset or new_value is not expr.value:
            result = expr.copy()
            result.base = new_base
            result.offset = new_offset
            result.value = new_value
            return result
        return expr



from __future__ import annotations
import logging

import pyvex

from angr.engines.engine import SimEngine
from angr.utils.constants import DEFAULT_STATEMENT

l = logging.getLogger(name=__name__)

# pylint:disable=arguments-differ,unused-argument,no-self-use


class VEXMixin(SimEngine):
    def __init__(self, project, **kwargs):
        super().__init__(project, **kwargs)
        self._vex_expr_handlers = []
        self._vex_stmt_handlers = []
        self.__init_handlers()

        self.irsb = None
        self.stmt_idx = None
        self.tmps = None

    def __init_handlers(self):
        self._vex_expr_handlers = [None] * pyvex.expr.tag_count
        self._vex_stmt_handlers = [None] * pyvex.stmt.tag_count
        for name, cls in vars(pyvex.expr).items():
            if isinstance(cls, type) and issubclass(cls, pyvex.expr.IRExpr) and cls is not pyvex.expr.IRExpr:
                self._vex_expr_handlers[cls.tag_int] = getattr(self, "_handle_vex_expr_" + name)
        for name, cls in vars(pyvex.stmt).items():
            if isinstance(cls, type) and issubclass(cls, pyvex.stmt.IRStmt) and cls is not pyvex.stmt.IRStmt:
                self._vex_stmt_handlers[cls.tag_int] = getattr(self, "_handle_vex_stmt_" + name)
        assert None not in self._vex_expr_handlers
        assert None not in self._vex_stmt_handlers

    def __getstate__(self):
        return (super().__getstate__(),)  # return unary tuple to not trip special behavior with falsey states

    def __setstate__(self, s):
        self.__init_handlers()
        super().__setstate__(s[0])

    # one size fits all?
    def _ty_to_bytes(self, ty):
        return pyvex.get_type_size(ty) // getattr(getattr(getattr(self, "state", None), "arch", None), "byte_width", 8)

    def _handle_vex_stmt(self, stmt: pyvex.stmt.IRStmt):
        handler = self._vex_stmt_handlers[stmt.tag_int]
        handler(stmt)

    def _handle_vex_expr(self, expr: pyvex.expr.IRExpr):
        handler = self._vex_expr_handlers[expr.tag_int]
        result = handler(expr)
        return self._instrument_vex_expr(result)

    def _instrument_vex_expr(self, result):
        return result

    def _handle_vex_const(self, const: pyvex.const.IRConst):
        return const.value

    #
    # Individual expression handlers go here
    #

    # expressions dependent on the state impl

    def _handle_vex_expr_RdTmp(self, expr: pyvex.expr.RdTmp):
        return self._perform_vex_expr_RdTmp(expr.tmp)

    def _perform_vex_expr_RdTmp(self, tmp):
        return self.tmps[tmp]

    def _handle_vex_expr_Get(self, expr: pyvex.expr.Get):
        return self._perform_vex_expr_Get(self._handle_vex_const(pyvex.const.U32(expr.offset)), expr.ty)

    def _perform_vex_expr_Get(self, offset, ty, **kwargs):
        return NotImplemented

    def _analyze_vex_expr_Load_addr(self, *a, **kw):
        return self._handle_vex_expr(*a, **kw)

    def _handle_vex_expr_Load(self, expr: pyvex.expr.Load):
        return self._perform_vex_expr_Load(self._analyze_vex_expr_Load_addr(expr.addr), expr.ty, expr.end)

    def _perform_vex_expr_Load(self, addr, ty, endness, **kwargs):
        return NotImplemented

    # expressions dependent on the data domain

    def _analyze_vex_expr_CCall_arg(self, *a, **kw):
        return self._handle_vex_expr(*a, **kw)

    def _handle_vex_expr_CCall(self, expr: pyvex.expr.CCall):
        return self._perform_vex_expr_CCall(
            expr.cee.name,
            expr.retty,
            [self._analyze_vex_expr_CCall_arg(arg) for arg in expr.args],
        )

    def _perform_vex_expr_CCall(self, func_name, ty, args, func=None):
        return NotImplemented

    def _handle_vex_expr_ITE(self, expr: pyvex.expr.ITE):
        return self._perform_vex_expr_ITE(
            self._handle_vex_expr(expr.cond), self._handle_vex_expr(expr.iftrue), self._handle_vex_expr(expr.iffalse)
        )

    def _perform_vex_expr_ITE(self, cond, ifTrue, ifFalse):
        return NotImplemented

    def _handle_vex_expr_Unop(self, expr: pyvex.expr.Unop):
        return self._handle_vex_expr_Op(expr)

    def _handle_vex_expr_Binop(self, expr: pyvex.expr.Unop):
        return self._handle_vex_expr_Op(expr)

    def _handle_vex_expr_Triop(self, expr: pyvex.expr.Unop):
        return self._handle_vex_expr_Op(expr)

    def _handle_vex_expr_Qop(self, expr: pyvex.expr.Unop):
        return self._handle_vex_expr_Op(expr)

    def _handle_vex_expr_Op(self, expr):
        return self._perform_vex_expr_Op(expr.op, [self._handle_vex_expr(arg) for arg in expr.args])

    def _perform_vex_expr_Op(self, op, args):
        return NotImplemented

    # fully implemented expressions

    def _handle_vex_expr_Const(self, expr: pyvex.expr.Const):
        return self._handle_vex_const(expr.con)

    def _analyze_vex_expr_GetI_ix(self, *a, **kw):
        return self._handle_vex_expr(*a, **kw)

    def _handle_vex_expr_GetI(self, expr: pyvex.expr.GetI):
        return self._perform_vex_expr_GetI(
            expr.descr.base,
            expr.descr.elemTy,
            expr.bias,
            self._analyze_vex_expr_GetI_ix(expr.ix),
            expr.descr.nElems,
        )

    def _perform_vex_expr_GetI_get(self, *a, **kw):
        return self._perform_vex_expr_Get(*a, **kw)

    def _perform_vex_expr_GetI(self, base, ty, bias, ix, nElems):
        offset = self._perform_vex_stmt_PutI_compute(base, ty, bias, ix, nElems)
        return self._perform_vex_expr_GetI_get(offset, ty)

    # oh boy.

    def _handle_vex_expr_GSPTR(self, expr: pyvex.expr.GSPTR):
        return NotImplemented

    def _handle_vex_expr_VECRET(self, expr: pyvex.expr.VECRET):
        return NotImplemented

    def _handle_vex_expr_Binder(self, expr: pyvex.expr.Binder):
        return NotImplemented

    #
    # Individual statement handlers go here
    #

    # stmt category 1: fluff

    def _handle_vex_stmt_IMark(self, stmt):
        pass

    def _handle_vex_stmt_NoOp(self, stmt):
        pass

    def _handle_vex_stmt_AbiHint(self, stmt):
        pass

    def _handle_vex_stmt_MBE(self, stmt):
        pass

    # stmt category 2: real shit

    def _analyze_vex_stmt_Put_data(self, *a, **kw):
        return self._handle_vex_expr(*a, **kw)

    def _handle_vex_stmt_Put(self, stmt):
        self._perform_vex_stmt_Put(
            self._handle_vex_const(pyvex.const.U32(stmt.offset)), self._analyze_vex_stmt_Put_data(stmt.data)
        )

    def _perform_vex_stmt_Put(self, offset, data, **kwargs):
        pass

    def _analyze_vex_stmt_WrTmp_data(self, *a, **kw):
        return self._handle_vex_expr(*a, **kw)

    def _handle_vex_stmt_WrTmp(self, stmt):
        self._perform_vex_stmt_WrTmp(stmt.tmp, self._analyze_vex_stmt_WrTmp_data(stmt.data))

    def _perform_vex_stmt_WrTmp(self, tmp, data):
        self.tmps[tmp] = data

    def _analyze_vex_stmt_Store_address(self, *a, **kw):
        return self._handle_vex_expr(*a, **kw)

    def _analyze_vex_stmt_Store_data(self, *a, **kw):
        return self._handle_vex_expr(*a, **kw)

    def _handle_vex_stmt_Store(self, stmt: pyvex.stmt.Store):
        self._perform_vex_stmt_Store(
            self._analyze_vex_stmt_Store_address(stmt.addr), self._analyze_vex_stmt_Store_data(stmt.data), stmt.end
        )

    def _perform_vex_stmt_Store(self, addr, data, endness, **kwargs):
        pass

    def _analyze_vex_stmt_Exit_guard(self, *a, **kw):
        return self._handle_vex_expr(*a, **kw)

    def _handle_vex_stmt_Exit(self, stmt: pyvex.stmt.Exit):
        self._perform_vex_stmt_Exit(
            self._analyze_vex_stmt_Exit_guard(stmt.guard), self._handle_vex_const(stmt.dst), stmt.jk
        )

    def _perform_vex_stmt_Exit(self, guard, target, jumpkind):
        pass

    def _analyze_vex_stmt_Dirty_arg(self, *a, **kw):
        return self._handle_vex_expr(*a, **kw)

    def _handle_vex_stmt_Dirty(self, stmt: pyvex.stmt.Dirty):
        return self._perform_vex_stmt_Dirty(
            stmt.cee.name,
            self.irsb.tyenv.lookup(stmt.tmp) if stmt.tmp not in (-1, 0xFFFFFFFF) else None,
            stmt.tmp,
            [self._analyze_vex_stmt_Dirty_arg(arg) for arg in stmt.args],
        )

    def _perform_vex_stmt_Dirty_wrtmp(self, *a, **kw):
        return self._perform_vex_stmt_WrTmp(*a, **kw)

    def _perform_vex_stmt_Dirty(self, func_name, ty, tmp, args):
        retval = self._perform_vex_stmt_Dirty_call(func_name, ty, args)
        if tmp not in (-1, 0xFFFFFFFF):
            self._perform_vex_stmt_Dirty_wrtmp(tmp, retval)

    def _perform_vex_stmt_Dirty_call(self, func_name, ty, args, func=None):
        return NotImplemented

    # stmt category 3: weird load/store patterns implemented in terms of above

    def _analyze_vex_stmt_PutI_ix(self, *a, **kw):
        return self._handle_vex_expr(*a, **kw)

    def _analyze_vex_stmt_PutI_data(self, *a, **kw):
        return self._handle_vex_expr(*a, **kw)

    def _handle_vex_stmt_PutI(self, stmt: pyvex.stmt.PutI):
        self._perform_vex_stmt_PutI(
            stmt.descr.base,
            stmt.descr.elemTy,
            stmt.bias,
            self._analyze_vex_stmt_PutI_ix(stmt.ix),
            stmt.descr.nElems,
            self._analyze_vex_stmt_PutI_data(stmt.data),
        )

    def _perform_vex_stmt_PutI_compute(self, base, elemTy, bias, ix, nElems):
        # base + ((bias + ix) % nElems) * elemSize
        elemSize = self._ty_to_bytes(elemTy)
        index = self._perform_vex_expr_Op("Iop_Add32", (self._handle_vex_const(pyvex.const.U32(bias)), ix))
        big_index = self._perform_vex_expr_Op("Iop_32HLto64", (self._handle_vex_const(pyvex.const.U32(0)), index))
        divmod_index = self._perform_vex_expr_Op(
            "Iop_DivModU64to32", (big_index, self._handle_vex_const(pyvex.const.U32(nElems)))
        )
        mod_index = self._perform_vex_expr_Op("Iop_64HIto32", (divmod_index,))
        offset = self._perform_vex_expr_Op("Iop_Mul32", (mod_index, self._handle_vex_const(pyvex.const.U32(elemSize))))
        return self._perform_vex_expr_Op("Iop_Add32", (self._handle_vex_const(pyvex.const.U32(base)), offset))

    def _perform_vex_stmt_PutI(self, base, elemSize, bias, ix, nElems, data):
        offset = self._perform_vex_stmt_PutI_compute(base, elemSize, bias, ix, nElems)
        self._perform_vex_stmt_Put(offset, data)

    def _analyze_vex_stmt_LLSC_addr(self, *a, **kw):
        return self._handle_vex_expr(*a, **kw)

    def _analyze_vex_stmt_LLSC_storedata(self, *a, **kw):
        return self._handle_vex_expr(*a, **kw)

    def _handle_vex_stmt_LLSC(self, stmt: pyvex.stmt.LLSC):
        self._perform_vex_stmt_LLSC(
            stmt.result,
            self._analyze_vex_stmt_LLSC_addr(stmt.addr),
            stmt.endness,
            self._analyze_vex_stmt_LLSC_storedata(stmt.storedata) if stmt.storedata is not None else None,
            self.irsb.tyenv.lookup(stmt.result),
        )

    def _perform_vex_stmt_LLSC_load(self, *a, **kw):
        return self._perform_vex_expr_Load(*a, **kw)

    def _perform_vex_stmt_LLSC_store(self, *a, **kw):
        return self._perform_vex_stmt_Store(*a, **kw)

    def _perform_vex_stmt_LLSC_wrtmp(self, *a, **kw):
        return self._perform_vex_stmt_WrTmp(*a, **kw)

    def _perform_vex_stmt_LLSC(self, result, addr, endness, storedata, ty):
        if storedata is None:
            load_result = self._perform_vex_stmt_LLSC_load(addr, ty, endness)
            self._perform_vex_stmt_LLSC_wrtmp(result, load_result)
        else:
            self._perform_vex_stmt_LLSC_store(addr, storedata, endness)
            self._perform_vex_stmt_LLSC_wrtmp(result, self._handle_vex_const(pyvex.const.U1(1)))

    def _analyze_vex_stmt_LoadG_addr(self, *a, **kw):
        return self._handle_vex_expr(*a, **kw)

    def _analyze_vex_stmt_LoadG_alt(self, *a, **kw):
        return self._handle_vex_expr(*a, **kw)

    def _analyze_vex_stmt_LoadG_guard(self, *a, **kw):
        return self._handle_vex_expr(*a, **kw)

    def _handle_vex_stmt_LoadG(self, stmt: pyvex.stmt.LoadG):
        self._perform_vex_stmt_LoadG(
            self._analyze_vex_stmt_LoadG_addr(stmt.addr),
            self._analyze_vex_stmt_LoadG_alt(stmt.alt),
            self._analyze_vex_stmt_LoadG_guard(stmt.guard),
            stmt.dst,
            stmt.cvt,
            stmt.end,
        )

    def _perform_vex_stmt_LoadG_load(self, *a, **kw):
        return self._perform_vex_expr_Load(*a, **kw)

    def _perform_vex_stmt_LoadG_widen(self, *a, **kw):
        return self._perform_vex_expr_Op(*a, **kw)

    def _perform_vex_stmt_LoadG_ite(self, *a, **kw):
        return self._perform_vex_expr_ITE(*a, **kw)

    def _perform_vex_stmt_LoadG_wrtmp(self, *a, **kw):
        return self._perform_vex_stmt_WrTmp(*a, **kw)

    def _perform_vex_stmt_LoadG_guard_condition(self, guard):
        return guard == 1

    def _perform_vex_stmt_LoadG(self, addr, alt, guard, dst, cvt, end):
        cvt_properties = {
            "ILGop_IdentV128": ("Ity_V128", None),  # 128 bit vector, no conversion */
            "ILGop_Ident64": ("Ity_I64", None),  # 64 bit, no conversion */
            "ILGop_Ident32": ("Ity_I32", None),  # 32 bit, no conversion */
            "ILGop_16Uto32": ("Ity_I16", "Iop_16Uto32"),  # 16 bit load, Z-widen to 32 */
            "ILGop_16Sto32": ("Ity_I16", "Iop_16Sto32"),  # 16 bit load, S-widen to 32 */
            "ILGop_8Uto32": ("Ity_I8", "Iop_8Uto32"),  # 8 bit load, Z-widen to 32 */
            "ILGop_8Sto32": ("Ity_I8", "Iop_8Sto32"),  # 8 bit load, S-widen to 32 */
        }

        # Because of how VEX's ARM lifter works, we may introduce non-existent register loads.
        # Here is an example:
        #
        # .text:0800408C ITTTT MI
        # .text:0800408E LDRMI   R2, =0x40020004
        # .text:08004090 LDRMI   R3
        #
        # 116 | ------ IMark(0x800408e, 2, 1) ------
        # 117 | t247 = Or32(t225,0x00000040)
        # 118 | t254 = armg_calculate_condition(t247,t227,t229,t231):Ity_I32
        # 119 | t262 = GET:I32(r2)
        # 120 | t263 = CmpNE32(t254,0x00000000)
        # 121 | t66 = if (t263) ILGop_Ident32(LDle(0x080040bc)) else t262
        # 122 | PUT(r2) = t66
        # 123 | PUT(pc) = 0x08004091
        # 124 | ------ IMark(0x8004090, 2, 1) ------
        # 125 | t280 = t263
        # 126 | t73 = if (t280) ILGop_Ident32(LDle(t66)) else t222
        #
        # t280 == t263 == the condition inside t66. Now t66 looks like this:
        #   <BV32 cond then 0x40020004 else reg_r2_861_32{UNINITIALIZED}>. since t280 is guarding the load from t66,
        # if the load from t66 is not aware of the condition that t280 is True, we will end up reading from r2_861_32,
        # which is not what the original instruction intended.
        # Therefore, the load from t66 should be aware of the condition that t280 is True. Or even better, don't
        # perform the read if the condition is evaluated to False.
        # We can perform another optimization: Let this condition be cond. When cond can be evaluated to either True or
        # False, we don't want to perform the read when the cond is the guard (which is a relatively cheap check) and
        # is False. When the cond is True, we perform the read with only the intended address (instead of the entire
        # guarded address). This way we get rid of the redundant load that should have existed in the first place.

        ty, cvt_op = cvt_properties[cvt]
        if self.state.solver.is_true(guard[0] == 0):
            self._perform_vex_stmt_LoadG_wrtmp(dst, alt)
            return
        load_result = self._perform_vex_stmt_LoadG_load(
            addr, ty, end, condition=self._perform_vex_stmt_LoadG_guard_condition(guard)
        )
        cvt_result = load_result if cvt_op is None else self._perform_vex_stmt_LoadG_widen(cvt_op, (load_result,))
        ite_result = self._perform_vex_stmt_LoadG_ite(guard, cvt_result, alt)
        self._perform_vex_stmt_LoadG_wrtmp(dst, ite_result)

    def _analyze_vex_stmt_StoreG_addr(self, *a, **kw):
        return self._handle_vex_expr(*a, **kw)

    def _analyze_vex_stmt_StoreG_data(self, *a, **kw):
        return self._handle_vex_expr(*a, **kw)

    def _analyze_vex_stmt_StoreG_guard(self, *a, **kw):
        return self._handle_vex_expr(*a, **kw)

    def _handle_vex_stmt_StoreG(self, stmt: pyvex.stmt.StoreG):
        self._perform_vex_stmt_StoreG(
            self._analyze_vex_stmt_StoreG_addr(stmt.addr),
            self._analyze_vex_stmt_StoreG_data(stmt.data),
            self._analyze_vex_stmt_StoreG_guard(stmt.guard),
            stmt.data.result_type(self.irsb.tyenv),
            stmt.end,
        )

    def _perform_vex_stmt_StoreG_load(self, *a, **kw):
        return self._perform_vex_expr_Load(*a, **kw)

    def _perform_vex_stmt_StoreG_ite(self, *a, **kw):
        return self._perform_vex_expr_ITE(*a, **kw)

    def _perform_vex_stmt_StoreG_store(self, *a, **kw):
        return self._perform_vex_stmt_Store(*a, **kw)

    def _perform_vex_stmt_StoreG_guard_condition(self, guard):
        return guard == 1

    def _perform_vex_stmt_StoreG(self, addr, data, guard, ty, endness, **kwargs):
        # perform the same optimization as in _perform_vex_stmt_LoadG
        if self.state.solver.is_true(guard[0] == 0):
            return
        self._perform_vex_stmt_StoreG_store(
            addr, data, endness, condition=self._perform_vex_stmt_StoreG_guard_condition(guard), **kwargs
        )

    def _analyze_vex_stmt_CAS_addr(self, *a, **kw):
        return self._handle_vex_expr(*a, **kw)

    def _analyze_vex_stmt_CAS_dataLo(self, *a, **kw):
        return self._handle_vex_expr(*a, **kw)

    def _analyze_vex_stmt_CAS_dataHi(self, *a, **kw):
        return self._handle_vex_expr(*a, **kw)

    def _analyze_vex_stmt_CAS_expdLo(self, *a, **kw):
        return self._handle_vex_expr(*a, **kw)

    def _analyze_vex_stmt_CAS_expdHi(self, *a, **kw):
        return self._handle_vex_expr(*a, **kw)

    def _handle_vex_stmt_CAS(self, stmt: pyvex.stmt.CAS):
        self._perform_vex_stmt_CAS(
            self._analyze_vex_stmt_CAS_addr(stmt.addr),
            self._analyze_vex_stmt_CAS_dataLo(stmt.dataLo),
            self._analyze_vex_stmt_CAS_dataHi(stmt.dataHi) if stmt.dataHi is not None else None,
            self._analyze_vex_stmt_CAS_expdLo(stmt.expdLo),
            self._analyze_vex_stmt_CAS_expdHi(stmt.expdHi) if stmt.expdHi is not None else None,
            stmt.oldLo,
            stmt.oldHi,
            stmt.endness,
            stmt.expdLo.result_type(self.irsb.tyenv),
        )

    def _perform_vex_stmt_CAS_load(self, *a, **kw):
        return self._perform_vex_expr_Load(*a, **kw)

    def _perform_vex_stmt_CAS_wrtmp(self, *a, **kw):
        return self._perform_vex_stmt_WrTmp(*a, **kw)

    def _perform_vex_stmt_CAS_cmp(self, *a, **kw):
        return self._perform_vex_expr_Op(*a, **kw)

    def _perform_vex_stmt_CAS_narrow(self, *a, **kw):
        return self._perform_vex_expr_Op(*a, **kw)

    def _perform_vex_stmt_CAS_widen(self, *a, **kw):
        return self._perform_vex_expr_Op(*a, **kw)

    def _perform_vex_stmt_CAS_storeg(self, *a, **kw):
        return self._perform_vex_stmt_StoreG(*a, **kw)

    def _perform_vex_stmt_CAS(self, addr, dataLo, dataHi, expdLo, expdHi, oldLo, oldHi, endness, ty):
        # - load mem
        # - compare
        # - store mem conditional
        # - store tmp
        double = dataHi is not None
        if double:
            ty, narrow_lo_op, narrow_hi_op, widen_op = {
                "Ity_I8": ("Ity_I16", "Iop_16to8", "Iop_16Hito8", "Iop_8HLto16"),
                "Ity_I16": ("Ity_I32", "Iop_32to16", "Iop_32HIto16", "Iop_16HLto32"),
                "Ity_I32": ("Ity_I64", "Iop_64to32", "Iop_64HIto32", "Iop_32HLto64"),
                "Ity_I64": ("Ity_V128", "Iop_128to64", "Iop_128HIto64", "Iop_64HLto128"),
            }[ty]
            data = self._perform_vex_stmt_CAS_widen(widen_op, (dataHi, dataLo))
            expd = self._perform_vex_stmt_CAS_widen(widen_op, (expdHi, expdLo))
        else:
            narrow_lo_op = narrow_hi_op = None
            data = dataLo
            expd = expdLo

        cmp_op = {
            "Ity_I8": "Iop_CmpEQ8",
            "Ity_I16": "Iop_CmpEQ16",
            "Ity_I32": "Iop_CmpEQ32",
            "Ity_I64": "Iop_CmpEQ64",
            "Ity_V128": "Iop_CmpEQ128",
        }[ty]

        val = self._perform_vex_stmt_CAS_load(addr, ty, endness)
        cmp = self._perform_vex_stmt_CAS_cmp(cmp_op, (val, expd))
        self._perform_vex_stmt_CAS_storeg(addr, data, cmp, ty, endness)

        if double:
            valHi = self._perform_vex_stmt_CAS_narrow(narrow_hi_op, (val,))
            valLo = self._perform_vex_stmt_CAS_narrow(narrow_lo_op, (val,))

            self._perform_vex_stmt_CAS_wrtmp(oldLo, valLo)
            self._perform_vex_stmt_CAS_wrtmp(oldHi, valHi)
        else:
            self._perform_vex_stmt_CAS_wrtmp(oldLo, val)

    #
    # block level handling
    #

    def _analyze_vex_defaultexit(self, *a, **kw):
        return self._handle_vex_expr(*a, **kw)

    def handle_vex_block(self, irsb: pyvex.IRSB):
        self.irsb = irsb
        self.tmps = [None] * self.irsb.tyenv.types_used

        for stmt_idx, stmt in enumerate(irsb.statements):
            self.stmt_idx = stmt_idx
            self._handle_vex_stmt(stmt)
        self.stmt_idx = DEFAULT_STATEMENT
        self._handle_vex_defaultexit(irsb.next, irsb.jumpkind)

    def _handle_vex_defaultexit(self, expr: pyvex.expr.IRExpr | None, jumpkind: str):
        self._perform_vex_defaultexit(self._analyze_vex_defaultexit(expr) if expr is not None else None, jumpkind)

    def _perform_vex_defaultexit(self, expr, jumpkind):
        pass



from __future__ import annotations

import argparse
import contextlib
import logging
import re
from typing import TYPE_CHECKING
from collections.abc import Generator

from rich import progress as rich_progress
from rich.logging import RichHandler
from rich.syntax import Syntax
from rich.console import Console
from rich.table import Column

import angr
from angr.analyses.decompiler import DECOMPILATION_PRESETS
from angr.analyses.decompiler.decompilation_options import PARAM_TO_OPTION
from angr.analyses.decompiler.structuring import STRUCTURER_CLASSES, DEFAULT_STRUCTURER
from angr.utils.formatting import ansi_color_enabled

if TYPE_CHECKING:
    from angr.knowledge_plugins.functions import Function

pdb = __import__("pdb")
with contextlib.suppress(ImportError):
    pdb = __import__("ipdb")


log = logging.getLogger(__name__)


NUMERIC_ARG_RE = re.compile(r"^(0x)?[a-fA-F0-9]+$")


def parse_function_args(proj: angr.Project, func_args: list[str] | None) -> Generator[Function]:
    """
    Generate a sequence of functions in the project kb by their identifier in func_args.

    :param proj:      Project to query.
    :param func_args: Sequence of function identifiers to query. None for all functions.
    """
    if func_args is None:
        yield from sorted(proj.kb.functions.values(), key=lambda f: f.addr)
        return

    for func_arg in func_args:
        if func_arg in proj.kb.functions:
            yield proj.kb.functions[func_arg]
            continue

        if NUMERIC_ARG_RE.match(func_arg):
            func_addr = int(func_arg, 0)
            if func_addr in proj.kb.functions:
                yield proj.kb.functions[func_addr]
                continue

        log.error('Function "%s" not found', func_arg)


def _make_status_console() -> tuple[Console, bool]:
    """Create a stderr console for status messages and determine if interactive status should be shown.

    Also installs a RichHandler on the angr logger so that log messages integrate
    properly with Rich's live progress displays instead of corrupting them.
    """
    console = Console(stderr=True)
    angr_logger = logging.getLogger("angr")
    angr_logger.handlers = [RichHandler(console=console, show_path=False, show_time=False)]
    angr_logger.propagate = False
    return console, console.is_terminal


def _make_progress(*extra_columns) -> tuple:
    """Build the standard column set for progress bars.

    The status text is placed *after* the bar so changing its length
    doesn't shift the bar and percentage around.
    """
    return (
        rich_progress.SpinnerColumn(),
        rich_progress.TextColumn("{task.description}"),
        rich_progress.BarColumn(),
        rich_progress.TaskProgressColumn(),
        rich_progress.TimeElapsedColumn(),
        *extra_columns,
    )


@contextlib.contextmanager
def _stderr_progress(console: Console, description: str, show: bool = True):
    """Show a progress bar on stderr for a long-running analysis.

    Yields a progress_callback compatible with angr's Analysis infrastructure.
    When not in a terminal, yields None (no progress display).
    """
    if not show:
        yield None
        return

    status_col = rich_progress.TextColumn("{task.fields[status]}")
    p = rich_progress.Progress(*_make_progress(status_col), console=console, redirect_stderr=True, transient=True)
    task_id = p.add_task(f"[bold]{description}", total=100, status="")

    def callback(percentage: float, text: str | None = None, **_kwargs) -> None:
        p.update(task_id, completed=percentage, status=text or "")

    with p:
        yield callback


class _NoOpTracker:
    """No-op tracker when progress display is disabled."""

    @contextlib.contextmanager
    def task(self, _name):  # pylint:disable=no-self-use
        yield None


_MAX_NAME_WIDTH = 20


@contextlib.contextmanager
def _multi_progress(console: Console, total: int, description: str, show: bool = True):
    """Progress tracker with an overall bar and per-item sub-task progress.

    The display is paused between tasks so stdout output doesn't corrupt the terminal.
    Yields a tracker whose ``task()`` context manager creates a sub-task progress bar
    and yields a ``progress_callback`` for the analysis.
    """
    if not show:
        yield _NoOpTracker()
        return

    prefix_len = 2 if total > 1 else 0  # "  " indent for sub-tasks
    desc_width = max(len(description), _MAX_NAME_WIDTH + prefix_len)
    status_col = rich_progress.TextColumn("{task.fields[status]}")
    p = rich_progress.Progress(
        rich_progress.SpinnerColumn(),
        rich_progress.TextColumn("{task.description}", table_column=Column(min_width=desc_width, max_width=desc_width)),
        rich_progress.BarColumn(),
        rich_progress.TaskProgressColumn(),
        rich_progress.TimeElapsedColumn(),
        status_col,
        console=console,
        redirect_stderr=True,
        transient=True,
    )
    overall = p.add_task(f"[bold]{description}", total=total, status="") if total > 1 else None

    class _Tracker:  # pylint:disable=missing-class-docstring
        @contextlib.contextmanager
        def task(self, name):  # pylint:disable=no-self-use
            prefix = "  " if overall is not None else ""
            display_name = (
                name
                if len(name) <= _MAX_NAME_WIDTH - len(prefix)
                else name[: _MAX_NAME_WIDTH - len(prefix) - 1] + "\u2026"
            )
            sub = p.add_task(f"{prefix}{display_name}", total=100, status="")
            p.start()

            def callback(percentage: float, text: str | None = None, **_kwargs) -> None:
                p.update(sub, completed=percentage, status=text or "")

            try:
                yield callback
            finally:
                p.remove_task(sub)
                if overall is not None:
                    p.advance(overall)
                p.stop()

    try:
        yield _Tracker()
    finally:
        p.stop()


def disassemble(args):
    """
    Disassemble functions.
    """
    err, show_status = _make_status_console()
    if not args.pbar:
        show_status = False

    loader_main_opts_kwargs = {}
    if args.blob:
        loader_main_opts_kwargs["backend"] = "blob"
    if args.base_addr is not None:
        loader_main_opts_kwargs["base_addr"] = args.base_addr
    if args.arch is not None:
        loader_main_opts_kwargs["arch"] = args.arch
    if args.entry_point is not None:
        loader_main_opts_kwargs["entry_point"] = args.entry_point

    proj = angr.Project(args.binary, auto_load_libs=False, main_opts=loader_main_opts_kwargs)

    with _stderr_progress(err, "Recovering control flow graph", show=show_status) as progress_cb:
        proj.analyses.CFG(  # pyright: ignore[reportCallIssue]
            normalize=True, data_references=True, progress_callback=progress_cb
        )

    # Collect disassemblable functions
    funcs = [
        f
        for f in parse_function_args(proj, args.functions)
        if not (f.is_plt or f.is_syscall or f.is_alignment or f.is_simprocedure)
    ]

    total = len(funcs)
    if total == 0:
        if show_status:
            err.print("[yellow]No disassemblable functions found[/yellow]")
        return

    success_count = 0
    error_count = 0

    with _multi_progress(err, total, "Disassembling", show=show_status) as tracker:
        for func in funcs:
            func_name = func.name or hex(func.addr)

            exception_string = ""
            text = ""
            with tracker.task(func_name):
                try:
                    disasm = proj.analyses.Disassembly(func)
                    text = disasm.render(show_bytes=True, min_edge_depth=10)
                except Exception as e:  # pylint:disable=broad-exception-caught
                    if args.pdb:
                        pdb.post_mortem(e.__traceback__)
                    elif not args.catch_exceptions:
                        raise
                    exception_string = str(e).replace("\n", " ")

            # Progress is paused here -- safe to print to stdout / stderr
            if exception_string:
                error_count += 1
                if show_status:
                    err.print(f"[red]Error disassembling {func_name}:[/red] {exception_string}")
                else:
                    log.error(exception_string)
            else:
                success_count += 1
                print(text)

    # Summary for interactive sessions
    if show_status and total > 1:
        if error_count:
            err.print(
                f"Disassembled {success_count}/{total} functions ({error_count} error{'s' if error_count != 1 else ''})"
            )
        else:
            err.print(f"Disassembled {success_count} functions")


def decompile(args):
    """
    Decompile functions.
    """
    structurer = args.structurer or DEFAULT_STRUCTURER.NAME
    should_highlight = ansi_color_enabled and not args.no_colors
    err, show_status = _make_status_console()
    if not args.pbar:
        show_status = False

    loader_main_opts_kwargs = {}
    if args.blob:
        loader_main_opts_kwargs["backend"] = "blob"
    if args.base_addr is not None:
        loader_main_opts_kwargs["base_addr"] = args.base_addr
    if args.arch is not None:
        loader_main_opts_kwargs["arch"] = args.arch
    if args.entry_point is not None:
        loader_main_opts_kwargs["entry_point"] = args.entry_point

    # Load binary
    proj = angr.Project(args.binary, auto_load_libs=False, main_opts=loader_main_opts_kwargs)

    # CFG recovery with progress on stderr
    with _stderr_progress(err, "Recovering control flow graph", show=show_status) as progress_cb:
        cfg = proj.analyses.CFG(  # pyright: ignore[reportCallIssue]
            normalize=True, data_references=True, progress_callback=progress_cb
        )

    # Complete calling conventions with progress on stderr
    if args.cca:
        with _stderr_progress(err, "Recovering calling conventions", show=show_status) as progress_cb:
            proj.analyses.CompleteCallingConventions(
                analyze_callsites=args.cca_callsites,
                progress_callback=progress_cb,  # pyright: ignore[reportCallIssue]
            )

    # Collect and normalize function identifiers
    if args.functions is None:
        func_ids = sorted(cfg.kb.functions)
    else:
        func_ids = []
        for func_arg in args.functions:
            try:
                func_id = int(func_arg, 0) if isinstance(func_arg, str) else func_arg
            except ValueError:
                func_id = func_arg
            if func_id not in cfg.functions:
                if args.catch_exceptions:
                    if show_status:
                        err.print(f"[yellow]Warning:[/yellow] Function {func_arg!r} not found")
                    else:
                        log.warning("Function %s does not exist in the CFG.", func_arg)
                    continue
                raise ValueError(f"Function {func_arg} does not exist in the CFG.")
            func_ids.append(func_id)

    # Filter to decompilable functions
    funcs = []
    for func_id in func_ids:
        f = cfg.functions[func_id]
        if f is not None and not (f.is_plt or f.is_syscall or f.is_alignment or f.is_simprocedure):
            funcs.append(f)

    total = len(funcs)
    if total == 0:
        if show_status:
            err.print("[yellow]No decompilable functions found[/yellow]")
        return

    # Prepare decompilation options
    dec_options = [
        (PARAM_TO_OPTION["structurer_cls"], structurer),
        (PARAM_TO_OPTION["show_casts"], not args.no_casts),
    ]
    if args.llm:
        dec_options.append(("llm_refine", True))

    out = Console(highlight=False)
    success_count = 0
    error_count = 0

    with _multi_progress(err, total, "Decompiling", show=show_status) as tracker:
        for func in funcs:
            func_name = func.name or hex(func.addr)

            # Decompile with progress bar showing per-stage status
            exception_string = ""
            dec = None
            with tracker.task(func_name) as progress_cb:
                try:
                    if args.catch_exceptions:
                        dec = proj.analyses.Decompiler(
                            func,
                            cfg=cfg,
                            options=dec_options,
                            preset=args.preset,
                            fail_fast=True,  # pyright: ignore[reportCallIssue]
                            progress_callback=progress_cb,  # pyright: ignore[reportCallIssue]
                        )
                    else:
                        dec = proj.analyses.Decompiler(
                            func,
                            cfg=cfg,
                            options=dec_options,
                            preset=args.preset,
                            progress_callback=progress_cb,  # pyright: ignore[reportCallIssue]
                        )
                except Exception as e:  # pylint:disable=broad-exception-caught
                    if args.pdb:
                        pdb.post_mortem(e.__traceback__)
                    elif not args.catch_exceptions:
                        raise
                    exception_string = str(e).replace("\n", " ")

            # Progress is paused here -- safe to print to stdout / stderr
            if not exception_string and (dec is None or not dec.codegen or not dec.codegen.text):
                exception_string = "Decompilation produced no output"

            if exception_string:
                error_count += 1
                if show_status:
                    err.print(f"[red]Error decompiling {func_name}:[/red] {exception_string}")
                else:
                    log.critical("Failed to decompile %s: %s", func_name, exception_string)
            else:
                success_count += 1
                assert dec is not None and dec.codegen is not None
                text = dec.codegen.text
                if should_highlight:
                    syntax = Syntax(text + "\n", "c", theme=args.theme, line_numbers=False)  # type: ignore[operator]
                    out.print(syntax)
                else:
                    print(text)

    # Trailing newline for plain text output
    if not should_highlight and success_count > 0:
        print()

    # Summary for interactive sessions
    if show_status and total > 1:
        if error_count:
            err.print(
                f"Decompiled {success_count}/{total} functions ({error_count} error{'s' if error_count != 1 else ''})"
            )
        else:
            err.print(f"Decompiled {success_count} functions")


def _add_common_args(subparser):
    """Add arguments common to all subcommands."""
    subparser.add_argument("binary", help="The path to the binary to analyze.")
    subparser.add_argument(
        "--catch-exceptions",
        help="""
        Catch exceptions during analysis. The scope of error handling may depend on the command used for analysis.
        If multiple functions are specified for analysis, each function will be handled individually.""",
        action="store_true",
        default=False,
    )
    subparser.add_argument(
        "--pdb",
        help="""
        Implies --catch-exceptions, and also launches a postmortem debug shell when we catch one.
        """,
        action="store_true",
        default=False,
    )
    subparser.add_argument(
        "--base-addr",
        help="""
        The base address of the binary. This is useful when the binary is loaded at a different address than the one
        specified in the ELF header.""",
        type=lambda x: int(x, 0),
        default=None,
    )
    subparser.add_argument(
        "--blob",
        help="Treat the input file as a raw binary blob instead of auto-detecting a file format.",
        action="store_true",
        default=False,
    )
    subparser.add_argument(
        "--arch",
        help="Architecture to use when loading a raw binary blob.",
        default=None,
    )
    subparser.add_argument(
        "--entry-point",
        help="Entry point to use when loading a raw binary blob.",
        type=lambda x: int(x, 0),
        default=None,
    )


def main():
    parser = argparse.ArgumentParser(description="The angr CLI allows you to decompile and analyze binaries.")
    parser.add_argument("--version", action="version", version=angr.__version__)
    parser.add_argument(
        "-v", "--verbose", action="count", default=0, help="Increase verbosity level (can be used multiple times)."
    )
    parser.add_argument(
        "-n",
        "--nopbar",
        action="store_false",
        dest="pbar",
        default=True,
        help="Disable progress bars; useful when debugging.",
    )
    subparsers = parser.add_subparsers(metavar="command", required=True)

    decompile_cmd_parser = subparsers.add_parser("decompile", aliases=["dec"], help=decompile.__doc__)
    decompile_cmd_parser.set_defaults(func=decompile)
    _add_common_args(decompile_cmd_parser)
    decompile_cmd_parser.add_argument(
        "--structurer",
        help="The structuring algorithm to use for decompilation.",
        choices=STRUCTURER_CLASSES.keys(),
        default=DEFAULT_STRUCTURER.NAME,
    )
    decompile_cmd_parser.add_argument(
        "--no-casts",
        help="Do not show type casts in the decompiled output.",
        action="store_true",
        default=False,
    )
    decompile_cmd_parser.add_argument(
        "--preset",
        help="The configuration preset to use for decompilation.",
        choices=DECOMPILATION_PRESETS,
        default="default",
    )
    decompile_cmd_parser.add_argument(
        "--cca",
        help="Enable full-binary function prototype recovery. Improves decompilation quality but may be slow on "
        "binaries with many functions.",
        action="store_true",
        default=False,
    )
    decompile_cmd_parser.add_argument(
        "--cca-callsites",
        help="When --cca (full-binary function prototype recovery) is enabled, also analyze call sites for better "
        "function prototype inference.",
        action="store_true",
        default=True,
    )
    decompile_cmd_parser.add_argument(
        "--llm",
        help="Use an LLM to refine the decompilation output. The LLM must be configured separately using environment "
        "variables. You may only need to set ANGR_LLM_MODEL (see pydantic-ai model list) and ANGR_LLM_API_KEY. See "
        "the documentation for angr.LLMClient for details in LLM configuration.",
        action="store_true",
        default=False,
    )
    decompile_cmd_parser.add_argument(
        "--functions",
        help="""
        The functions to decompile. Functions can either be expressed as names found in the
        symbols of the binary or as addresses like: 0x401000.""",
        nargs="+",
    )
    decompile_cmd_parser.add_argument(
        "--no-colors",
        help="Disable syntax highlighting in the decompiled output.",
        action="store_true",
        default=False,
    )
    decompile_cmd_parser.add_argument(
        "--theme",
        help="The syntax highlighting theme to use (only if rich is installed and colors are enabled).",
        default="ansi_dark",
    )

    disassemble_cmd_parser = subparsers.add_parser("disassemble", aliases=["dis"], help=disassemble.__doc__)
    disassemble_cmd_parser.set_defaults(func=disassemble)
    _add_common_args(disassemble_cmd_parser)
    disassemble_cmd_parser.add_argument(
        "--functions",
        help="""
        The functions to disassemble. Functions can either be expressed as names found in the
        symbols of the binary or as addresses like: 0x401000.""",
        nargs="+",
    )

    args = parser.parse_args()

    if args.blob:
        if args.arch is None:
            parser.error("--blob requires --arch")
        if args.base_addr is None:
            parser.error("--blob requires --base-addr")
        if args.entry_point is None:
            parser.error("--blob requires --entry-point")

    log_level = max(logging.ERROR - (10 * args.verbose), logging.DEBUG)
    logging.getLogger("angr").setLevel(log_level)

    args.func(args)


if __name__ == "__main__":
    main()



from __future__ import annotations

from enum import Enum


class StackItemType(Enum):
    """
    Enum for the type of stack items.
    """

    UNKNOWN = 0
    SAVED_BP = 1
    SAVED_REGS = 2
    ARGUMENT = 3
    RET_ADDR = 4
    STACK_CANARY = 5


class StackItem:
    """
    A stack item describes a piece of data that is stored on the stack at a certain offset (usually negative).
    """

    offset: int
    size: int
    name: str
    item_type: StackItemType

    def __init__(self, offset: int, size: int, name: str, item_type: StackItemType = StackItemType.UNKNOWN):
        self.offset = offset
        self.size = size
        self.name = name
        self.item_type = item_type

    def __repr__(self):
        return f"<StackItem {self.name} {self.item_type!s} at {self.offset:#x} ({self.size}b)>"



from __future__ import annotations
from typing import TYPE_CHECKING

from io import BytesIO

from cle.backends import Blob

from angr.knowledge_base import KnowledgeBase
from .simos import SimOS

if TYPE_CHECKING:
    from angr import Project


class SimSnimmucNxp(SimOS):
    """
    This class implements the "OS" for a bare-metal firmware used at an imaginary company.
    """

    def __init__(self, project: Project, name=None, **kwargs):  # pylint:disable=unused-argument
        super().__init__(project, name=name)

    def configure_project(self):
        # pattern match the entry point to figure out if we support parsing this binary
        entry_bytes = self.project.loader.memory.load(self.project.entry, 3 * 4)
        if entry_bytes != b"\x94!\xff\xf0|\x08\x02\xa6\x90\x01\x00\x14":
            return

        entry_block = self.project.factory.block(self.project.entry)
        try:
            first_sync = next(
                iter(
                    [
                        idx
                        for idx, insn in enumerate(entry_block.disassembly.insns)
                        if insn.mnemonic in {"sync", "isync"}
                    ]
                )
            )
        except StopIteration:
            return

        # run this block and acquire initial registers for each function
        state = self.project.factory.blank_state(addr=self.project.entry)
        # set garbage value to key registers
        key_registers = ["r13", "r2", "r14", "r15", "r16"]
        GARBAGE = 0xDEADBEEF
        for key_reg in key_registers:
            setattr(state.regs, "_" + key_reg, GARBAGE)
        simgr = self.project.factory.simgr(state)
        simgr.step(num_inst=first_sync)
        if simgr.active and len(simgr.active) == 1:
            stepped_state = simgr.one_active
        else:
            return

        reg_values = {}
        for key_reg in key_registers:
            reg_values[key_reg] = getattr(stepped_state.regs, "_" + key_reg).concrete_value
            if reg_values[key_reg] in {None, GARBAGE}:
                # umm the register is not initialized. unsupported?
                return

        # TODO: Make them part of the ABI
        self.function_initial_registers = reg_values

        # load SDATA, SDATA2, and a few other regions
        mappings = {}

        # this is just CRAZY...
        # TODO: Better resilience
        tmp_kb = KnowledgeBase(self.project)
        self.project.analyses.CFG(
            regions=[(self.project.entry, self.project.entry + 180)], data_references=False, kb=tmp_kb
        )
        # take the last function
        func = tmp_kb.functions[self.project.entry]
        second_to_last_block = sorted(func.blocks, key=lambda x: x.addr)[-2]
        if second_to_last_block.vex.jumpkind != "Ijk_Call":
            return
        init_func_addr = second_to_last_block.vex.next
        if not isinstance(init_func_addr, int):
            return

        # lift one block
        init_func_block = self.project.factory.block(init_func_addr)
        if init_func_block.vex.jumpkind != "Ijk_Call":
            return

        section_init_func_addr = init_func_block.vex.next
        if not isinstance(section_init_func_addr, int):
            return

        self.project.analyses.CFG(
            regions=[(section_init_func_addr, section_init_func_addr + 0x324)], data_references=False, kb=tmp_kb
        )
        section_init_func = tmp_kb.functions[section_init_func_addr]

        sorted_blocks = sorted(section_init_func.blocks, key=lambda x: x.addr)
        sdata_section_init_call = sorted_blocks[25]
        if sdata_section_init_call.vex.jumpkind != "Ijk_Call":
            return
        sdata_section_init_func = sdata_section_init_call.vex.next
        if not isinstance(sdata_section_init_func, int):
            return

        # more pattern matching
        state = self.project.factory.blank_state(addr=sdata_section_init_func)
        for key_reg in ["r28", "r29", "r30"]:
            setattr(state.regs, "_" + key_reg, GARBAGE)
        simgr = self.project.factory.simgr(state)
        simgr.step()
        if simgr.active and len(simgr.active) == 1:
            stepped_state = simgr.one_active
        else:
            return

        sdata_reg_values = {}
        for key_reg in ["r28", "r29", "r30"]:
            sdata_reg_values[key_reg] = getattr(stepped_state.regs, "_" + key_reg).concrete_value
            if sdata_reg_values[key_reg] in {None, GARBAGE}:
                # umm the register is not initialized. unsupported?
                return

        mappings[sdata_reg_values["r30"]] = (sdata_reg_values["r29"], sdata_reg_values["r28"] - sdata_reg_values["r29"])

        # TODO: Implement support for SDATA2 and other sections
        # mappings = {
        #     0x60005850: (0x30A734, 0x60008D30 - 0x60005850),
        #     0x60011DA0: (0x3165EC, 0x60014580 - 0x60011DA0),
        #     0x60014580: (0x32AC48, 0x60061638 - 0x60014580),
        # }

        for mem_base, (source_addr, size) in mappings.items():
            backing = BytesIO()
            backing.write(self.project.loader.memory.load(source_addr, size))
            backing.seek(0)

            blob = Blob(
                binary=None,
                binary_stream=backing,
                base_addr=mem_base,
                offset=0,
                arch=self.project.arch,
            )
            self.project.loader.dynamic_load(blob)

        # FIXME: Use ret_offset from the calling convention
        self.project.arch.ret_offset = self.project.arch.registers["r3"][0]



from __future__ import annotations
import os
import importlib
import importlib.machinery
import importlib.util
import logging
from collections.abc import Callable

l = logging.getLogger(name=__name__)


def auto_import_packages(base_module, base_path, ignore_dirs=(), ignore_files=(), scan_modules=True):
    for lib_module_name in os.listdir(base_path):
        if lib_module_name in ignore_dirs or lib_module_name == "__pycache__":
            continue

        lib_path = os.path.join(base_path, lib_module_name)
        init_path = os.path.join(lib_path, "__init__.py")
        if not os.path.isfile(init_path):
            l.debug("Not a module: %s", lib_module_name)
            continue

        l.debug("Loading %s.%s", base_module, lib_module_name)

        try:
            package = importlib.import_module(f".{lib_module_name}", base_module)
        except ImportError:
            l.warning("Unable to autoimport package %s.%s", base_module, lib_module_name, exc_info=True)
        else:
            if scan_modules:
                for name, mod in auto_import_modules(
                    f"{base_module}.{lib_module_name}", lib_path, ignore_files=ignore_files
                ):
                    if name not in dir(package):
                        setattr(package, name, mod)
            yield lib_module_name, package


def auto_import_modules(base_module, base_path, ignore_files=(), filter_func: Callable | None = None):
    for proc_file_name in os.listdir(base_path):
        if not proc_file_name.endswith(".py"):
            continue
        if proc_file_name in ignore_files or proc_file_name == "__init__.py":
            continue
        proc_module_name = proc_file_name[:-3]
        if filter_func is not None and not filter_func(proc_module_name):
            continue

        try:
            proc_module = importlib.import_module(f".{proc_module_name}", base_module)
        except ImportError:
            l.warning("Unable to autoimport module %s.%s", base_module, proc_module_name, exc_info=True)
            continue
        else:
            yield proc_module_name, proc_module


def filter_module(mod, type_req=None, subclass_req=None):
    for name in dir(mod):
        val = getattr(mod, name)
        if type_req is not None and not isinstance(val, type_req):
            continue
        if subclass_req is not None and not issubclass(val, subclass_req):
            continue
        yield name, val


def auto_import_source_files(base_path, ignore_files=(), filter_func: Callable | None = None):
    for proc_file_name in os.listdir(base_path):
        if not proc_file_name.endswith(".py"):
            continue
        if proc_file_name in ignore_files or proc_file_name == "__init__.py":
            continue

        proc_module_name = proc_file_name[:-3]
        if filter_func is not None and not filter_func(proc_file_name):
            continue

        proc_path = os.path.join(base_path, proc_file_name)
        try:
            loader = importlib.machinery.SourceFileLoader(proc_module_name, proc_path)
            spec = importlib.util.spec_from_loader(proc_module_name, loader)
            proc_module = importlib.util.module_from_spec(spec)
            loader.exec_module(proc_module)
        except ImportError:
            l.warning("Unable to autoimport module from file %s", proc_path, exc_info=True)
            continue
        else:
            yield proc_module_name, proc_module



from __future__ import annotations

from angr.ailment.expression import BinaryOp, Convert, Const, ITE, Extract, Insert
from .base import PeepholeOptimizationExprBase

_MASKS = {
    1: 1,
    8: 0xFF,
    16: 0xFFFF,
    32: 0xFFFFFFFF,
    64: 0xFFFFFFFF_FFFFFFFF,
}


class RemoveRedundantBitmasks(PeepholeOptimizationExprBase):
    """
    Remove redundant bitmasking operations.
    """

    __slots__ = ()

    NAME = "Remove redundant bitmasks"
    expr_classes = (BinaryOp, Convert, Extract, Insert)

    def optimize(self, expr: BinaryOp | Convert | Extract | Insert, **kwargs):
        if isinstance(expr, BinaryOp):
            return self._optimize_BinaryOp(expr)
        if isinstance(expr, Convert):
            return RemoveRedundantBitmasks._optimize_Convert(expr)
        if isinstance(expr, Extract):
            return self._optimize_Extract(expr)
        if isinstance(expr, Insert):
            return self._optimize_Insert(expr)
        return None

    def _optimize_Insert(self, expr: Insert):
        if (
            self.project is not None
            and isinstance(expr.base, BinaryOp)
            and expr.base.op == "And"
            and isinstance((mask := expr.base.operands[1]), Const)
            and isinstance(mask.value, int)
            and isinstance(expr.offset, Const)
            and isinstance(expr.offset.value, int)
            # is this correct for big-endian??
            and _MASKS.get(expr.value.bits, 0) << (expr.offset.value * self.project.arch.byte_width) == mask
        ):
            # Insert(v0 & mask, offset, v1) where mask/offset guarantee
            # that the only bits we get from v0 will just be replaced with v1
            return Insert(expr.idx, Const(None, None, 0, expr.bits), expr.offset, expr.value, expr.endness, **expr.tags)
        return None

    def _optimize_Extract(self, expr: Extract):
        if (
            self.project is not None
            and isinstance(expr.base, BinaryOp)
            and expr.base.op == "And"
            and isinstance((mask := expr.base.operands[1]), Const)
            and isinstance(mask.value, int)
            and _MASKS.get(expr.bits) == mask.value
            and expr.is_lsb_extract()
        ):
            return Convert(expr.idx, expr.base.bits, expr.bits, False, expr.base, **expr.tags)
        return None

    def _optimize_BinaryOp(self, expr: BinaryOp):
        # And(expr, full_N_bitmask) ==> expr
        # And(SHR(expr, N), bitmask)) ==> SHR(expr, N)
        # And(Div(Conv(M->N, expr), P), 2 ** M - 1) ==> Div(Conv(M->N, expr), P)  where M < N
        # And(Conv(1->N, expr), bitmask) ==> Conv(1->N, expr)
        # And(Conv(1->N, bool_expr), bitmask) ==> Conv(1->N, bool_expr)
        # And(ITE(?, const_expr, const_expr), bitmask) ==> ITE(?, const_expr, const_expr)
        if expr.op == "And" and isinstance(expr.operands[1], Const):
            inner_expr = expr.operands[0]
            if expr.operands[1].value == _MASKS.get(inner_expr.bits):
                return inner_expr

            if isinstance(inner_expr, BinaryOp):
                if inner_expr.op == "Shr":
                    mask = expr.operands[1]
                    shift_val = inner_expr.operands[1]
                    if (
                        isinstance(shift_val, Const)
                        and shift_val.value in _MASKS
                        and mask.value == _MASKS.get(int(64 - shift_val.value))
                    ):
                        return inner_expr
                if inner_expr.op == "Div" and isinstance(inner_expr.operands[0], Convert):
                    from_bits = inner_expr.operands[0].from_bits
                    to_bits = inner_expr.operands[0].to_bits
                    if from_bits < to_bits:
                        mask = expr.operands[1]
                        if mask.value == _MASKS.get(from_bits):
                            return inner_expr

            if isinstance(inner_expr, Convert) and self.is_bool_expr(inner_expr.operand):
                # useless masking
                return inner_expr
            if (
                isinstance(expr.operands[0], ITE)
                and isinstance(expr.operands[0].iftrue, Const)
                and isinstance(expr.operands[0].iffalse, Const)
            ):
                # is the masking unnecessary?
                mask = expr.operands[1].value
                ite = expr.operands[0]
                if mask == 0xFF and ite.iftrue.value <= 0xFF and ite.iffalse.value <= 0xFF:
                    # yes!
                    return ite

        return None

    @staticmethod
    def _optimize_Convert(expr: Convert):
        # Conv(64->32, (expr & bitmask) + expr)
        # => Conv(64->32, (expr + expr))
        if (
            expr.op == "Convert"
            and expr.from_bits > expr.to_bits
            and isinstance(expr.operand, BinaryOp)
            and expr.operand.op == "Add"
        ):
            operand_expr = expr.operand
            op0, op1 = operand_expr.operands
            if (
                isinstance(op0, BinaryOp)
                and op0.op == "And"
                and isinstance(op0.operands[1], Const)
                and op0.operands[1].value == _MASKS.get(expr.to_bits)
            ):
                new_op0 = op0.operands[0]
                replaced, new_operand_expr = operand_expr.replace(op0, new_op0)
                if replaced:
                    expr.operand = new_operand_expr
                    return expr
        # Conv(64->32, (expr) - (expr) & 0xffffffff<64>)))
        # => Conv(64->32, (expr - expr))
        elif (
            expr.op == "Convert"
            and expr.from_bits > expr.to_bits
            and isinstance(expr.operand, BinaryOp)
            and expr.operand.op == "Sub"
        ):
            operand_expr = expr.operand
            op0, op1 = operand_expr.operands
            if (
                isinstance(op1, BinaryOp)
                and op1.op == "And"
                and isinstance(op1.operands[1], Const)
                and op1.operands[1].value == _MASKS.get(expr.to_bits)
            ):
                new_op1 = op1.operands[0]
                replaced, new_operand_expr = operand_expr.replace(op1, new_op1)
                if replaced:
                    expr.operand = new_operand_expr
                    return expr

        return None



from __future__ import annotations
from collections import defaultdict

import angr.ailment as ailment

from angr.analyses.decompiler.sequence_walker import SequenceWalker
from angr.analyses.decompiler.structuring.structurer_nodes import LoopNode


class ControlFlowStructureCounter(SequenceWalker):
    """
    Counts the number of different types of control flow structures found in a sequence of nodes.
    This should be used after the sequence has been simplified.
    """

    def __init__(self, node):
        handlers = {
            LoopNode: self._handle_Loop,
            ailment.Block: self._handle_Block,
        }
        super().__init__(handlers)

        self.while_loops = 0
        self.do_while_loops = 0
        self.for_loops = 0
        self.ordered_labels = []
        self.goto_targets = defaultdict(int)

        self.walk(node)

        # eliminate gotos without labels
        self.goto_targets = {k: v for k, v in self.goto_targets.items() if k in self.ordered_labels}
        # correct labels that are not used
        self.ordered_labels = [lbl for lbl in self.ordered_labels if lbl in self.goto_targets]

    # pylint: disable=unused-argument
    def _handle_Block(self, node: ailment.Block, **kwargs):
        if not node.statements:
            return

        for stmt in node.statements:
            # labels found in the block at this point will be labels in the output
            if isinstance(stmt, ailment.statement.Label):
                label_addr = stmt.tags.get("ins_addr")
                if label_addr is not None:
                    self.ordered_labels.append(label_addr)

            # goto targets found in the block at this point will be goto targets in the output
            if isinstance(stmt, ailment.statement.Jump):
                target = stmt.target
                target_value = target.value if target is not None and isinstance(target, ailment.Expr.Const) else None
                if target_value is not None:
                    self.goto_targets[target_value] += 1

    def _handle_Loop(self, node: LoopNode, **kwargs):
        if node.sort == "while":
            self.while_loops += 1
        elif node.sort == "do-while":
            self.do_while_loops += 1
        elif node.sort == "for":
            self.for_loops += 1

        return super()._handle_Loop(node, **kwargs)



from __future__ import annotations
import logging

from angr.sim_procedure import SimProcedure

l = logging.getLogger(name=__name__)


class DataGraphError(Exception):
    pass


class DataGraphMeta:
    def __init__(self):
        self._p = None

    def _irsb(self, in_state):
        """
        We expect a VSA state here.
        """
        return self._p.factory.successors(in_state)

    def _vfg_node(self, addr):
        """
        Gets vfg node at @addr
        Returns VFGNode or None
        """
        for n in self._vfg._nodes.values():
            if n.addr == addr:
                return n
        raise DataGraphError(f"No VFG node at 0x{addr:x}")

    def get_irsb_at(self, addr):
        n = self._vfg_node(addr)
        if n is None:
            raise DataGraphError("No VFG node at this address")
        return self._irsb(n.state)

    def pp(self, imarks=False):
        """
        Pretty print the graph.  @imarks determine whether the printed graph
        represents instructions (coarse grained) for easier navigation,  or
        exact statements.
        """
        for e in self.graph.edges():
            data = dict(self.graph.get_edge_data(e[0], e[1]))
            data["label"] = str(data["label"]) + " ; " + self._simproc_info(e[0]) + self._simproc_info(e[1])
            self._print_edge(e, data, imarks)

    def _print_edge(self, e, data, imarks=False):
        pp = []
        for stmt in e:
            # true case is a SimProcedure
            s = f"(0x{stmt[0]:x}, {stmt[1]})" if imarks is False or stmt[1] == -1 else f"[0x{self._imarks[stmt]:x}]"
            pp.append(s)

        print(pp[0] + " -> " + pp[1] + " : " + str(data))

    def _branch(self, live_defs, node, path=""):
        """
        Recursive function, it branches in every possible path in the VFG.
        @live_defs: a dict {addr:stmt} of live definitions at the start point
        @node: the starting vfg node

        Returns: the address of the block where the execution stops
        """

        irsb = self._irsb(node.state)
        path = path + " -> " + hex(irsb.addr)

        if isinstance(irsb, SimProcedure):
            self._simproc_map[irsb.addr] = repr(irsb)

        l.debug(f"--> Branch: running block 0x{irsb.addr:x}")
        block = self._make_block(irsb, live_defs)
        self._imarks.update(block._imarks)
        if block.stop is True:
            # l.debug(" ### Stopping at block 0x%x" % (irsb.addr))
            l.debug(f" ### End of path {path}")
            return irsb.addr
        succ = self._vfg._graph.successors(node)

        defer = []
        for s in succ:
            # Consider fake returns last
            if self._vfg._graph.edge[node][s]["jumpkind"] == "Ijk_FakeRet":
                defer.append(s)
                continue
            # We need to make a copy of the dict !
            self._branch(dict(block.live_defs), s, path)

            # We explore every other paths before taking fake rets.
            # Ideally, we want to take fake rets only when functions don't
            # return.
            for s in defer:
                self._branch(dict(block.live_defs), s, path)
        return None

    def _make_block(self, vfg_node, live_defs):
        raise DataGraphError("Not Implemented")

    def _simproc_info(self, node):
        if node[0] in self._simproc_map:
            return self._simproc_map[node[0]]
        return ""



from __future__ import annotations
from typing import Any

import networkx

from angr.analyses.decompiler.utils import structured_node_is_simple_return
from angr.utils.graph import PostDominators, TemporaryNode
from .phoenix import PhoenixStructurer


class SAILRStructurer(PhoenixStructurer):
    """
    The SAILR structuring algorithm is the phoenix-based algorithm from the USENIX 2024 paper SAILR.
    The entirety of the algorithm is implemented across this class and various optimization passes in the decompiler.
    To find each optimization class, simply search for optimizations which reference this class.NAME.

    At a high-level, SAILR does three things different from the traditional Phoenix schema-based algorithm:
    1. It recursively structures the graph, rather than doing it in a single pass. This allows decisions to be made
        based on the current state of what the decompilation would look like.
    2. It performs deoptimizations targeting specific optimizations that introduces gotos and mis-structured code.
        It can only do this because of the recursive nature of the algorithm.
    3. It uses a more advanced heuristic for virtualizing edges, which is implemented in this class.

    Additionally, some changes in Phoenix are only activated when SAILR is used.
    """

    NAME = "sailr"

    def __init__(self, region, improve_phoenix=True, postdom_max_edges=10, postdom_max_graph_size=50, **kwargs):
        self._postdom_max_edges = postdom_max_edges
        self._postdom_max_graph_size = postdom_max_graph_size
        super().__init__(
            region,
            improve_algorithm=improve_phoenix,
            **kwargs,
        )

    def _order_virtualizable_edges(self, graph: networkx.DiGraph, edges: list, node_seq: dict[Any, int]) -> list:
        """
        The criteria for "best" is defined by a variety of heuristics described below.
        """
        if len(edges) <= 1:
            return edges

        # TODO: the graph we have here is not an accurate graph and can have no "entry node". We need a better graph.
        try:
            entry_node = next(node for node in graph.nodes if graph.in_degree(node) == 0)
        except StopIteration:
            entry_node = None

        edges = sorted(edges, key=lambda edge: (edge[0].addr, edge[1].addr))
        best_edges = edges
        if entry_node is not None:
            # collect sibling counts for edges
            edge_sibling_count = {}
            for edge in edges:
                _, dst = edge
                sibling_cnt = graph.in_degree[dst] - 1
                if sibling_cnt == 0:
                    continue
                edge_sibling_count[edge] = sibling_cnt

            # H1: the edge that has the least amount of sibling edges should be virtualized first
            # this is believed to reduce the amount of virtualization needed in future rounds and increase
            # the edges that enter a single outer-scope if-stmt
            if edge_sibling_count:
                min_sibling_count = min(edge_sibling_count.values())
                best_edges = [edge for edge, cnt in edge_sibling_count.items() if cnt == min_sibling_count]
                if len(best_edges) == 1:
                    return best_edges

            if len(edges) <= self._postdom_max_edges and len(graph) <= self._postdom_max_graph_size:
                # dominator analysis is expensive, so we only do it for small graphs
                edge_postdom_count = {}
                graph_copy = networkx.DiGraph(graph)
                for edge in edges:
                    _, dst = edge
                    graph_copy.remove_edge(*edge)
                    post_dom_graph = PostDominators(graph_copy, entry_node).post_dom
                    post_doms = set()
                    for postdom_node, dominatee in post_dom_graph.edges():
                        if not isinstance(postdom_node, TemporaryNode) and not isinstance(dominatee, TemporaryNode):
                            post_doms.add((postdom_node, dominatee))
                    edge_postdom_count[edge] = len(post_doms)
                    # add back the edge
                    graph_copy.add_edge(*edge)

                # create the next heuristic based on the best edges from the previous heuristic
                filtered_edge_postdom_count = edge_postdom_count.copy()
                for edge in list(edge_postdom_count.keys()):
                    if edge not in best_edges:
                        del filtered_edge_postdom_count[edge]
                if filtered_edge_postdom_count:
                    edge_postdom_count = filtered_edge_postdom_count

                # H2: the edge, when removed, that causes the most post-dominators of the graph should be virtualized
                # first. this is believed to make the code more linear looking be reducing the amount of scopes.
                # informally, we believe post-dominators to be an inverse indicator of the number of scopes present
                if edge_postdom_count:
                    max_postdom_count = max(edge_postdom_count.values())
                    best_edges = [edge for edge, cnt in edge_postdom_count.items() if cnt == max_postdom_count]
                    if len(best_edges) == 1:
                        return best_edges

            # H3: the edge that goes directly to a return statement should be virtualized first
            # this is believed to be good because it can be corrected in later optimization by duplicating
            # the return
            candidate_edges = best_edges
            best_edges = []
            for src, dst in candidate_edges:
                if graph.has_node(dst) and structured_node_is_simple_return(dst, graph):
                    best_edges.append((src, dst))

            if len(best_edges) == 1:
                return best_edges
            if not best_edges:
                best_edges = candidate_edges

        # if we have another tie, or we never used improved heuristics, then we do the default ordering.
        return super()._order_virtualizable_edges(graph, best_edges, node_seq)



# pylint:disable=too-many-boolean-expressions
from __future__ import annotations
from typing import TYPE_CHECKING, TypeAlias
from collections import defaultdict

import pyvex

from angr.utils.bits import u2s
from angr.block import Block
from angr.analyses.analysis import Analysis
from angr.analyses import AnalysesHub
from angr.knowledge_plugins.functions import Function
from angr.codenode import BlockNode, HookNode, FuncNode
from angr.engines.light import SimEngineNostmtVEX, SimEngineLight
from angr.calling_conventions import SimRegArg, SimStackArg, default_cc
from angr.sim_type import SimTypeBottom, SimTypeFunction
from angr.utils.types import dereference_simtype_by_lib
from .utils import is_sane_register_variable

if TYPE_CHECKING:
    from angr.codenode import CodeNode

# if you're going to change these to an enum, please do some benchmarking
# (kind, subkind, offset)
KIND_SP = 0
KIND_REG = 1
KIND_STACKVAL = 2
KIND_CONST = 3

# for KIND_SP
SUBKIND_SP = 0
SUBKIND_BP = 1

# for KIND_REG, subkind is reg offset
# for KIND_STACKVAL subkind is source stack offset
# offset is const offset from original value, or value for KIND_CONST

FactData: TypeAlias = tuple[int, int, int] | None


class FactCollectorState:
    """
    The abstract state for FactCollector.
    """

    __slots__ = (
        "bp_value",
        "callee_stored_regs",
        "ins_addr",
        "pointer_arg_derefs",
        "reg_reads",
        "reg_reads_count",
        "reg_writes",
        "simple_regs",
        "simple_stack",
        "sp_value",
        "stack_reads",
        "stack_writes",
        "tmps",
    )

    def __init__(self):
        self.tmps: dict[int, FactData] = {}
        self.simple_stack: dict[int, FactData] = {}
        self.simple_regs: dict[int, FactData] = {}
        self.ins_addr = 0

        self.callee_stored_regs: dict[int, int] = {}  # reg offset -> stack offset
        self.reg_reads = {}
        self.reg_reads_count = defaultdict(int)
        self.reg_writes: set[int] = set()
        self.stack_reads = {}
        self.stack_writes: set[int] = set()
        self.pointer_arg_derefs: defaultdict[FactData, int] = defaultdict(int)
        self.sp_value = 0
        self.bp_value = 0

    def register_read(self, offset: int, size_in_bytes: int):
        self.reg_reads_count[offset] += 1
        if offset in self.reg_writes:
            return
        if offset not in self.reg_reads:
            self.reg_reads[offset] = size_in_bytes
        else:
            self.reg_reads[offset] = max(self.reg_reads[offset], size_in_bytes)

    def register_read_undo(self, offset: int) -> None:
        if offset not in self.reg_reads or offset not in self.reg_reads_count:
            return
        self.reg_reads_count[offset] -= 1
        if self.reg_reads_count[offset] == 0:
            self.reg_reads.pop(offset)
            self.reg_reads_count.pop(offset)

    def register_written(self, offset: int, size_in_bytes: int):
        for o in range(size_in_bytes):
            self.reg_writes.add(offset + o)

    def stack_read(self, offset: int, size_in_bytes: int):
        if offset in self.stack_writes:
            return
        if offset not in self.stack_reads:
            self.stack_reads[offset] = size_in_bytes
        else:
            self.stack_reads[offset] = max(self.stack_reads[offset], size_in_bytes)

    def stack_written(self, offset: int, size_int_bytes: int):
        for o in range(size_int_bytes):
            self.stack_writes.add(offset + o)

    def copy(self, with_tmps: bool = True) -> FactCollectorState:
        new_state = FactCollectorState()
        new_state.reg_reads = self.reg_reads.copy()
        new_state.stack_reads = self.stack_reads.copy()
        new_state.stack_writes = self.stack_writes.copy()
        new_state.reg_writes = self.reg_writes.copy()
        new_state.callee_stored_regs = self.callee_stored_regs.copy()
        new_state.sp_value = self.sp_value
        new_state.bp_value = self.bp_value
        new_state.simple_stack = self.simple_stack.copy()
        new_state.simple_regs = self.simple_regs.copy()
        new_state.reg_reads_count = self.reg_reads_count.copy()
        new_state.pointer_arg_derefs = self.pointer_arg_derefs.copy()
        new_state.ins_addr = self.ins_addr
        if with_tmps:
            new_state.tmps = self.tmps.copy()
        return new_state


binop_handler = SimEngineNostmtVEX[FactCollectorState, FactData, FactCollectorState].binop_handler


class SimEngineFactCollectorVEX(
    SimEngineNostmtVEX[FactCollectorState, FactData, None],
    SimEngineLight[FactCollectorState, FactData, Block, None],
):
    """
    The engine for FactCollector.
    """

    def __init__(self, project, bp_as_gpr: bool, track_arg_uses: bool, seen_reg_uses: defaultdict[int, int]):
        self.bp_as_gpr = bp_as_gpr
        self.track_arg_uses = track_arg_uses
        self.seen_reg_uses = seen_reg_uses
        super().__init__(project)

    def _process_block_end(self, stmt_result: list, whitelist: set[int] | None) -> None:
        if self.block.vex.jumpkind == "Ijk_Call" and self.arch.ret_offset is not None:
            self.state.register_written(self.arch.ret_offset, self.arch.bytes)

    def _top(self, bits: int):
        return None

    def _is_top(self, expr) -> bool:
        return expr is None

    def _expr(self, expr):
        r = super()._expr(expr)
        if (
            r is not None
            and r[0] == KIND_REG
            and not (
                isinstance((stmt := self.block.vex.statements[self.stmt_idx]), pyvex.stmt.WrTmp) and stmt.data is expr
            )
        ):
            # don't count wrtmp datas
            self.seen_reg_uses[r[1]] += 1
        return r

    def _handle_conversion(self, from_size: int, to_size: int, signed: bool, operand: pyvex.expr.IRExpr):
        return None

    def _handle_stmt_IMark(self, stmt: pyvex.stmt.IMark):
        self.state.ins_addr = stmt.addr

    def _handle_stmt_Put(self, stmt):
        v = self._expr(stmt.data)
        # there are cases like  VMOV.F32        S0, S0
        # so we need to check if this register write is actually a no-op
        if isinstance(stmt.data, pyvex.IRExpr.RdTmp):
            t = self.state.tmps.get(stmt.data.tmp, None)
            if t is not None and t[0] == KIND_REG and t[1] == stmt.offset:
                same_ins_read = False
                for i in range(self.stmt_idx, -1, -1):
                    if i >= self.block.vex.stmts_used:
                        break
                    prev_stmt = self.block.vex.statements[i]
                    if isinstance(prev_stmt, pyvex.IRStmt.IMark):
                        break
                    if (
                        isinstance(prev_stmt, pyvex.IRStmt.WrTmp)
                        and prev_stmt.tmp == stmt.data.tmp
                        and isinstance(prev_stmt.data, pyvex.IRExpr.Get)
                        and prev_stmt.data.offset == stmt.offset
                    ):
                        same_ins_read = True
                        break
                if same_ins_read:
                    # we need to revert the read operation as well
                    self.state.register_read_undo(stmt.offset)
                return

        if stmt.offset == self.arch.sp_offset and v is not None and v[0] == KIND_SP:
            self.state.sp_value = v[2]
        elif stmt.offset == self.arch.bp_offset and v is not None and v[1] == KIND_SP:
            self.state.bp_value = v[2]
        else:
            self.state.register_written(stmt.offset, stmt.data.result_size(self.tyenv) // self.arch.byte_width)
            self.state.simple_regs[stmt.offset] = v

    def _handle_stmt_Store(self, stmt: pyvex.IRStmt.Store):
        addr = self._expr(stmt.addr)
        data = self._expr(stmt.data)
        if addr is None or not (addr[0] == KIND_SP or (addr[0] in (KIND_REG, KIND_STACKVAL) and self.track_arg_uses)):
            return

        if addr[0] == KIND_SP:
            self.state.stack_written(addr[2], stmt.data.result_size(self.tyenv) // self.arch.byte_width)
            if data is not None and data[0] == KIND_REG and data[2] == 0:
                # push reg; we record the stored register as well as the stack slot offset
                self.state.callee_stored_regs[data[1]] = u2s(addr[2], self.arch.bits)
            self.state.simple_stack[addr[2]] = data
        else:
            self.state.pointer_arg_derefs[addr] |= 2

    def _handle_stmt_WrTmp(self, stmt: pyvex.IRStmt.WrTmp):
        v = self._expr(stmt.data)
        self.state.tmps[stmt.tmp] = v

    def _handle_expr_Const(self, expr: pyvex.IRExpr.Const):
        return (KIND_CONST, 0, expr.con.value)

    def _handle_expr_GSPTR(self, expr):
        return (KIND_CONST, 0, 0)

    def _handle_expr_Get(self, expr):
        if expr.offset == self.arch.sp_offset:
            return (KIND_SP, 0, self.state.sp_value)
        if expr.offset == self.arch.bp_offset and not self.bp_as_gpr:
            return (KIND_SP, 0, self.state.bp_value)
        bits = expr.result_size(self.tyenv)
        self.state.register_read(expr.offset, bits // self.arch.byte_width)
        return self.state.simple_regs.get(expr.offset, (KIND_REG, expr.offset, 0))

    def _handle_expr_GetI(self, expr):
        return None

    def _handle_expr_ITE(self, expr):
        return None

    def _handle_expr_Load(self, expr):
        addr = self._expr(expr.addr)
        if addr is None or not (addr[0] == KIND_SP or (addr[0] in (KIND_REG, KIND_STACKVAL) and self.track_arg_uses)):
            return None

        if addr[0] == KIND_SP:
            self.state.stack_read(addr[2], expr.result_size(self.tyenv) // self.arch.byte_width)
            return self.state.simple_stack.get(addr[2], (KIND_STACKVAL, addr[2], 0))

        self.state.pointer_arg_derefs[addr] |= 1
        return None

    def _handle_expr_RdTmp(self, expr):
        return self.state.tmps.get(expr.tmp, None)

    def _handle_expr_VECRET(self, expr):
        return None

    @binop_handler
    def _handle_binop_Add(self, expr):
        op0, op1 = self._expr(expr.args[0]), self._expr(expr.args[1])
        if op0 is None or op1 is None:
            return None
        if op0[0] == KIND_CONST:
            return (op1[0], op1[1], op1[2] + op0[2])
        if op1[0] == KIND_CONST:
            return (op0[0], op0[1], op0[2] + op1[2])
        return None

    @binop_handler
    def _handle_binop_Sub(self, expr):
        op0, op1 = self._expr(expr.args[0]), self._expr(expr.args[1])
        if op0 is None or op1 is None:
            return None
        if op0[0] == KIND_CONST:
            return (op1[0], op1[1], op1[2] - op0[2])
        if op1[0] == KIND_CONST:
            return (op0[0], op0[1], op0[2] - op1[2])
        return None

    @binop_handler
    def _handle_binop_And(self, expr):
        op0, op1 = self._expr(expr.args[0]), self._expr(expr.args[1])
        if op0 is not None and op0[0] == KIND_SP:
            return op0
        if op1 is not None and op1[0] == KIND_SP:
            return op1
        return None


class FactCollector(Analysis):
    """
    An extremely fast analysis that extracts necessary facts of a function for CallingConventionAnalysis to make
    decision on the calling convention and prototype of a function.
    """

    def __init__(
        self, func: Function, max_depth: int = 100, track_arg_uses: bool = False, track_arg_passthru: bool = False
    ):
        self.function = func
        self._max_depth = max_depth
        self._track_arg_uses = track_arg_uses
        self._track_arg_passthru = track_arg_passthru
        self.callsites: dict[int, tuple[Function, list[FactData]]] = {}

        self.input_args: list[SimRegArg | SimStackArg] | None = None
        self.unused_args: list[SimRegArg] = []
        self.retval_size: int | None = None
        self.pointer_arg_derefs: defaultdict[FactData, int] = defaultdict(int)
        self._seen_reg_uses: defaultdict[int, int] = defaultdict(int)

        self._analyze()

    def _analyze(self):
        # breadth-first search using function graph, collect registers and stack variables that are written to as well
        # as read from, until max_depth is reached

        end_states = self._analyze_startpoint()
        self._analyze_endpoints_for_retval_size(end_states)
        callee_restored_regs = self._analyze_endpoints_for_restored_regs()
        self._determine_input_args(end_states, callee_restored_regs)

    def _analyze_startpoint(self) -> list[FactCollectorState]:
        func_graph = self.function.transition_graph
        startpoint = self.function.startpoint
        if startpoint is None:
            return []

        bp_as_gpr = self.function.info.get("bp_as_gpr", False)
        engine = SimEngineFactCollectorVEX(self.project, bp_as_gpr, self._track_arg_uses, self._seen_reg_uses)
        init_state = FactCollectorState()
        if self.project.arch.call_pushes_ret:
            init_state.sp_value = self.project.arch.bytes
        init_state.bp_value = init_state.sp_value

        traversed = set()
        queue: list[
            tuple[
                int,
                FactCollectorState,
                CodeNode | BlockNode | HookNode | FuncNode,
                BlockNode | HookNode | FuncNode | None,
            ]
        ] = [(0, init_state, startpoint, None)]
        end_states: list[FactCollectorState] = []
        while queue:
            depth, state, node, retnode = queue.pop(0)
            if isinstance(node, BlockNode) and node in traversed:
                continue
            traversed.add(node)

            if depth > self._max_depth:
                end_states.append(state)
                break

            if isinstance(node, BlockNode) and node.size == 0:
                continue
            func: Function | None = None
            if isinstance(node, (HookNode, FuncNode)):
                # attempt to convert it into a function
                if self.kb.functions.contains_addr(node.addr):
                    func = self.kb.functions.get_by_addr(node.addr)
                else:
                    continue
            if func is not None:
                if func.calling_convention is not None and func.prototype is not None:
                    # consume args and overwrite the return register
                    self._handle_function(state, func)
                if func.returning is False or retnode is None:
                    # the function call does not return
                    end_states.append(state)
                else:
                    # enqueue the retnode, but we don't increment the depth
                    new_state = state.copy()
                    if self.project.arch.call_pushes_ret and not func.is_syscall:
                        new_state.sp_value += self.project.arch.bytes
                    queue.append((depth, new_state, retnode, None))
                continue

            block = self.project.factory.block(node.addr, size=node.size)
            engine.process(state, block=block)

            successor_added = False
            call_succ, ret_succ = None, None
            for _, succ, data in func_graph.out_edges(node, data=True):
                edge_type = data.get("type")
                outside = data.get("outside", False)
                if depth + 1 <= self._max_depth:
                    if edge_type == "fake_return":
                        if succ not in traversed:
                            ret_succ = succ
                    elif edge_type == "transition" and not outside:
                        if succ not in traversed:
                            successor_added = True
                            queue.append((depth + 1, state.copy(), succ, None))
                    elif edge_type in {"call", "syscall"} or (edge_type == "transition" and outside):
                        # a call or a tail-call
                        # note that it's ok to traverse a called function multiple times
                        if not isinstance(succ, FuncNode) and not self.kb.functions.contains_addr(succ.addr):
                            # not sure who we are calling
                            continue
                        call_succ = succ
            if call_succ is not None:
                successor_added = True
                queue.append((depth + 1, state.copy(), call_succ, ret_succ))

            if not successor_added:
                end_states.append(state)

        return end_states

    def _handle_function(self, state: FactCollectorState, func: Function) -> None:
        try:
            if func.calling_convention is not None and func.prototype is not None:
                func_prototype = (
                    dereference_simtype_by_lib(func.prototype, func.prototype_libname)
                    if func.prototype_libname is not None
                    else func.prototype
                )
                arg_locs = func.calling_convention.arg_locs(func_prototype)
            else:
                return
        except (TypeError, ValueError):
            return

        if None in arg_locs:
            return

        if self._track_arg_passthru:
            self.callsites[state.ins_addr] = (func, [])
        for arg_loc in arg_locs:
            val: FactData = None
            for loc in arg_loc.get_footprint():
                if isinstance(loc, SimRegArg):
                    base_offset = self.project.arch.registers[loc.reg_name][0]
                    state.register_read(base_offset + loc.reg_offset, loc.size)
                    if self._track_arg_passthru:
                        val = state.simple_regs.get(base_offset, (KIND_REG, base_offset, 0))
                elif isinstance(loc, SimStackArg):
                    sp_value = state.sp_value
                    if sp_value is not None:
                        offset = sp_value + loc.stack_offset
                        state.stack_read(offset, loc.size)
                        if self._track_arg_passthru:
                            val = state.simple_stack.get(offset, (KIND_STACKVAL, offset, 0))
            if self._track_arg_passthru:
                if val is not None and val[0] == KIND_REG:
                    self._seen_reg_uses[val[1]] += 1
                self.callsites[state.ins_addr][1].append(val)

        # clobber caller-saved regs
        for reg_name in func.calling_convention.CALLER_SAVED_REGS:
            offset = self.project.arch.registers[reg_name][0]
            state.register_written(offset, self.project.arch.registers[reg_name][1])
            state.simple_regs[offset] = None

    def _analyze_endpoints_for_retval_size(self, end_states):
        """
        Analyze all endpoints to determine the return value size.
        """
        func_graph = self.function.transition_graph
        cc_cls = default_cc(
            self.project.arch.name, platform=self.project.simos.name if self.project.simos is not None else None
        )
        if cc_cls is None:
            # don't know what the calling convention may be... give up
            return
        cc = cc_cls(self.project.arch)
        if isinstance(cc.RETURN_VAL, SimRegArg):
            retreg_offset = cc.RETURN_VAL.check_offset(self.project.arch)
        else:
            return

        retval_sizes = []
        for endpoint in self.function.endpoints:
            assert isinstance(endpoint, (BlockNode, HookNode))
            traversed = set()
            queue: list[tuple[int, CodeNode]] = [(0, endpoint)]
            while queue:
                depth, node = queue.pop(0)
                if isinstance(node, BlockNode) and node in traversed:
                    continue
                traversed.add(node)

                if depth > 3:
                    break

                if isinstance(node, BlockNode) and node.size == 0:
                    continue

                func = None
                if isinstance(node, (FuncNode, HookNode)):
                    # attempt to convert it into a function
                    if self.kb.functions.contains_addr(node.addr):
                        func = self.kb.functions.get_by_addr(node.addr)
                    else:
                        continue
                if func is not None:
                    if (
                        func.calling_convention is not None
                        and func.prototype is not None
                        and func.prototype.returnty is not None
                        and not isinstance(func.prototype.returnty, SimTypeBottom)
                    ):
                        # assume the function overwrites the return variable
                        returnty_size = func.prototype.returnty.with_arch(self.project.arch).size
                        assert returnty_size is not None
                        retval_size = returnty_size // self.project.arch.byte_width
                        retval_sizes.append(retval_size)
                    continue

                # if this block ends with a call to a function, we process the function first
                func_succs = [
                    succ
                    for succ in func_graph.successors(node)
                    if isinstance(succ, (FuncNode, HookNode)) or self.kb.functions.contains_addr(succ.addr)
                ]
                if len(func_succs) == 1:
                    succ = func_succs[0]
                    func_succ: Function | None = None
                    if isinstance(succ, (BlockNode, HookNode, FuncNode)) and self.kb.functions.contains_addr(succ.addr):
                        # attempt to convert it into a function
                        func_succ = self.kb.functions.get_by_addr(succ.addr)
                    if func_succ is not None and func_succ.name != "_security_check_cookie":
                        if (
                            func_succ.calling_convention is not None
                            and func_succ.prototype is not None
                            and func_succ.prototype.returnty is not None
                            and not isinstance(func_succ.prototype.returnty, SimTypeBottom)
                        ):
                            # assume the function overwrites the return variable
                            proto = func_succ.prototype
                            if func_succ.prototype_libname is not None:
                                # we need to deref the prototype in case it uses SimTypeRef internally
                                proto = dereference_simtype_by_lib(proto, func_succ.prototype_libname)

                            assert isinstance(proto, SimTypeFunction) and proto.returnty is not None
                            returnty_size = proto.returnty.with_arch(self.project.arch).size
                            if returnty_size is None:
                                # it may be None if somehow we cannot resolve a SimTypeRef; we fall back to the full
                                # machine word size
                                retval_size = self.project.arch.bytes
                            else:
                                retval_size = returnty_size // self.project.arch.byte_width
                            retval_sizes.append(retval_size)
                            continue
                        if (
                            func_succ.prototype is not None
                            and func_succ.prototype.returnty is not None
                            and isinstance(func_succ.prototype.returnty, SimTypeBottom)
                        ):
                            # callee is void - don't scan VEX for return values since the call
                            # just clobbers rax without returning anything meaningful
                            continue

                block = self.project.factory.block(node.addr, size=node.size)

                # collect tmps so we can trace back through RdTmp
                tmp_definitions = {}
                for stmt in block.vex.statements:
                    if isinstance(stmt, pyvex.IRStmt.WrTmp):
                        tmp_definitions[stmt.tmp] = stmt.data

                # scan the block statements backwards to find writes to the return value register
                retval_size = None
                for stmt in reversed(block.vex.statements):
                    if isinstance(stmt, pyvex.IRStmt.Put):
                        assert block.vex.tyenv is not None
                        size = stmt.data.result_size(block.vex.tyenv) // self.project.arch.byte_width

                        # check if this 64-bit write is actually a sign/zero-extended 32-bit value.
                        if size == 8 and self.project.arch.bits == 64:
                            expr = stmt.data

                            if isinstance(expr, pyvex.IRExpr.RdTmp):
                                expr = tmp_definitions.get(expr.tmp, expr)

                            if isinstance(expr, pyvex.IRExpr.Unop) and expr.op in {"Iop_32Sto64", "Iop_32Uto64"}:
                                size = 4

                        if stmt.offset == retreg_offset:
                            retval_size = max(size, 1)

                if retval_size is not None:
                    retval_sizes.append(retval_size)
                    continue

                for pred, _, data in func_graph.in_edges(node, data=True):
                    edge_type = data.get("type")
                    if pred not in traversed and depth + 1 <= self._max_depth:
                        if edge_type in {"call", "syscall"}:
                            continue
                        if edge_type in {"transition", "fake_return"}:
                            queue.append((depth + 1, pred))

        # ARM/AArch64: R0/X0 used for both arg0 and return
        if not retval_sizes:
            first_arg_offset = None
            if cc.ARG_REGS:
                arg0_name = cc.ARG_REGS[0]
                if arg0_name in self.project.arch.registers:
                    first_arg_offset = self.project.arch.registers[arg0_name][0]

            if first_arg_offset is not None and first_arg_offset == retreg_offset:
                is_written = False
                for state in end_states:
                    if retreg_offset in state.reg_writes:
                        is_written = True
                        break

                if not is_written:
                    retval_sizes.append(self.project.arch.bytes)

        self.retval_size = max(retval_sizes) if retval_sizes else None

    def _analyze_endpoints_for_restored_regs(self):
        """
        Analyze all endpoints to determine the restored registers.
        """
        func_graph = self.function.transition_graph
        callee_restored_regs = set()

        sp_masks = {
            0xFFFFFFFE,
            0xFFFFFFFC,
            0xFFFFFFF8,
            0xFFFFFFF0,
            0xFFFFFFFF_FFFFFFFE,
            0xFFFFFFFF_FFFFFFFC,
            0xFFFFFFFF_FFFFFFF8,
            0xFFFFFFFF_FFFFFFF0,
        }
        for endpoint in self.function.endpoints:
            assert isinstance(endpoint, (BlockNode, HookNode))
            traversed = set()
            queue: list[tuple[int, CodeNode]] = [(0, endpoint)]
            while queue:
                depth, node = queue.pop(0)
                traversed.add(node)

                if depth > 3:
                    break

                if isinstance(node, BlockNode) and node.size == 0:
                    continue
                if isinstance(node, (HookNode, FuncNode)):
                    continue

                block = self.project.factory.block(node.addr, size=node.size)
                # scan the block statements backwards to find all statements that restore registers from the stack
                tmps = {}
                for stmt in block.vex.statements:
                    if isinstance(stmt, pyvex.IRStmt.WrTmp):
                        if isinstance(stmt.data, pyvex.IRExpr.Get) and stmt.data.offset in {
                            self.project.arch.bp_offset,
                            self.project.arch.sp_offset,
                        }:
                            tmps[stmt.tmp] = "sp"
                        elif (
                            isinstance(stmt.data, pyvex.IRExpr.Load)
                            and isinstance(stmt.data.addr, pyvex.IRExpr.RdTmp)
                            and tmps.get(stmt.data.addr.tmp) == "sp"
                        ):
                            tmps[stmt.tmp] = "stack_value"
                        elif isinstance(stmt.data, pyvex.IRExpr.Const):
                            tmps[stmt.tmp] = "const"
                        elif isinstance(stmt.data, pyvex.IRExpr.Binop):
                            if stmt.data.op.startswith("Iop_Add") or stmt.data.op.startswith("Iop_Sub"):
                                if (
                                    isinstance(stmt.data.args[0], pyvex.IRExpr.RdTmp)
                                    and tmps.get(stmt.data.args[0].tmp) == "sp"
                                ) or (
                                    isinstance(stmt.data.args[1], pyvex.IRExpr.RdTmp)
                                    and tmps.get(stmt.data.args[1].tmp) == "sp"
                                ):
                                    tmps[stmt.tmp] = "sp"
                            elif stmt.data.op.startswith("Iop_And"):  # noqa: SIM102
                                if (
                                    isinstance(stmt.data.args[0], pyvex.IRExpr.RdTmp)
                                    and tmps.get(stmt.data.args[0].tmp) == "sp"
                                    and isinstance(stmt.data.args[1], pyvex.IRExpr.Const)
                                    and stmt.data.args[1].con.value in sp_masks
                                ) or (
                                    isinstance(stmt.data.args[1], pyvex.IRExpr.RdTmp)
                                    and tmps.get(stmt.data.args[1].tmp) == "sp"
                                    and isinstance(stmt.data.args[0], pyvex.IRExpr.Const)
                                    and stmt.data.args[0].con.value in sp_masks
                                ):
                                    tmps[stmt.tmp] = "sp"
                    if isinstance(stmt, pyvex.IRStmt.Put):
                        assert block.vex.tyenv is not None
                        size = stmt.data.result_size(block.vex.tyenv) // self.project.arch.byte_width
                        # is the data loaded from the stack?
                        if (
                            size == self.project.arch.bytes
                            and isinstance(stmt.data, pyvex.IRExpr.RdTmp)
                            and tmps.get(stmt.data.tmp) == "stack_value"
                        ):
                            callee_restored_regs.add(stmt.offset)

                for pred, _, data in func_graph.in_edges(node, data=True):
                    edge_type = data.get("type")
                    if pred not in traversed and depth + 1 <= self._max_depth and edge_type == "transition":
                        queue.append((depth + 1, pred))

        # remove offsets of registers that may store return values from callee_restored_regs
        ret_reg_offsets = set()
        cc_cls = default_cc(
            self.project.arch.name, platform=self.project.simos.name if self.project.simos is not None else None
        )
        if cc_cls is not None:
            cc = cc_cls(self.project.arch)
            if isinstance(cc.RETURN_VAL, SimRegArg):
                retreg_offset = cc.RETURN_VAL.check_offset(self.project.arch)
                ret_reg_offsets.add(retreg_offset)
            if isinstance(cc.OVERFLOW_RETURN_VAL, SimRegArg):
                retreg_offset = cc.OVERFLOW_RETURN_VAL.check_offset(self.project.arch)
                ret_reg_offsets.add(retreg_offset)
            if isinstance(cc.FP_RETURN_VAL, SimRegArg):
                try:
                    retreg_offset = cc.FP_RETURN_VAL.check_offset(self.project.arch)
                    ret_reg_offsets.add(retreg_offset)
                except KeyError:
                    # register name does not exist
                    pass

        return callee_restored_regs.difference(ret_reg_offsets)

    def _determine_input_args(self, end_states: list[FactCollectorState], callee_restored_regs: set[int]) -> None:
        self.input_args = []
        reg_offset_created = set()
        callee_saved_regs = set()
        callee_saved_reg_stack_offsets = set()

        if self._track_arg_uses:
            for state in end_states:
                for k, v in state.pointer_arg_derefs.items():
                    self.pointer_arg_derefs[k] |= v

        # determine callee-saved registers
        unused_hint_offsets = set()
        for state in end_states:
            for reg_offset, stack_offset in state.callee_stored_regs.items():
                if reg_offset in callee_restored_regs:
                    callee_saved_regs.add(reg_offset)
                    callee_saved_reg_stack_offsets.add(stack_offset)
                elif self._seen_reg_uses[reg_offset] < 2:
                    unused_hint_offsets.add(reg_offset)

        for state in end_states:
            for offset, size in state.reg_reads.items():
                if (
                    offset in reg_offset_created
                    or offset == self.project.arch.bp_offset
                    or not is_sane_register_variable(self.project.arch, offset, size)
                    or offset in callee_saved_regs
                ):
                    continue
                reg_offset_created.add(offset)
                reg_name = self.project.arch.translate_register_name(offset, size=size)
                arg = SimRegArg(reg_name, size)
                self.input_args.append(arg)
                if offset in unused_hint_offsets:
                    self.unused_args.append(arg)

        stack_offset_created = set()
        ret_addr_offset = 0 if not self.project.arch.call_pushes_ret else self.project.arch.bytes
        # handle shadow stack args
        cc_cls = default_cc(
            self.project.arch.name, platform=self.project.simos.name if self.project.simos is not None else None
        )
        stackarg_sp_buff = cc_cls.STACKARG_SP_BUFF if cc_cls is not None else 0
        for state in end_states:
            for offset, size in state.stack_reads.items():
                offset = u2s(offset, self.project.arch.bits)
                if offset - ret_addr_offset > stackarg_sp_buff:
                    if offset in stack_offset_created or offset in callee_saved_reg_stack_offsets:
                        continue
                    stack_offset_created.add(offset)
                    arg = SimStackArg(offset - ret_addr_offset, size)
                    self.input_args.append(arg)


AnalysesHub.register_default("FunctionFactCollector", FactCollector)



from __future__ import annotations
from collections.abc import Iterable

import claripy
import logging

from angr.calling_conventions import DEFAULT_CC, default_cc, SimRegArg
from angr.engines.successors import SuccessorsEngine, SimSuccessors
from angr.misc.ux import once
from angr.utils.constants import DEFAULT_STATEMENT
from angr import sim_options as o
from angr import errors
from .lifter import PcodeLifterEngineMixin, IRSB
from .emulate import PcodeEmulatorMixin

l = logging.getLogger(__name__)

# pylint:disable=abstract-method


class HeavyPcodeMixin(
    SuccessorsEngine,
    PcodeLifterEngineMixin,
    PcodeEmulatorMixin,
):
    """
    Execution engine based on P-code, Ghidra's IR.

    Responds to the following parameters to the step stack:

    - irsb:        The P-Code IRSB object to use for execution. If not provided one will be lifted.
    - skip_stmts:  The number of statements to skip in processing
    - last_stmt:   Do not execute any statements after this statement
    - thumb:       Whether the block should be force to be lifted in ARM's THUMB mode. (FIXME)
    - extra_stop_points:
                   An extra set of points at which to break basic blocks
    - insn_bytes:  A string of bytes to use for the block instead of the project.
    - size:        The maximum size of the block, in bytes.
    - num_inst:    The maximum number of instructions.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._addr = None
        self._insn_bytes = None
        self._thumb = None
        self._size = None
        self._num_inst = None
        self._extra_stop_points = None

    def process_successors(
        self,
        successors: SimSuccessors,
        irsb: IRSB | None = None,
        insn_bytes: bytes | None = None,
        thumb: bool = False,
        size: int | None = None,
        num_inst: int | None = None,
        extra_stop_points: Iterable[int] | None = None,
        **kwargs,
    ) -> None:
        # pylint:disable=arguments-differ
        if type(successors.addr) is not int:
            return super().process_successors(
                successors,
                extra_stop_points=extra_stop_points,
                num_inst=num_inst,
                size=size,
                insn_bytes=insn_bytes,
                **kwargs,
            )

        successors.sort = "IRSB"
        successors.description = "IRSB"
        self.state.history.recent_block_count = 1
        self.state.scratch.guard = claripy.true()
        self.state.scratch.sim_procedure = None
        addr = successors.addr
        self.state.scratch.bbl_addr = addr
        self.state.scratch.irsb = irsb

        self._addr = addr
        self._insn_bytes = insn_bytes
        self._thumb = thumb
        self._size = size
        self._num_inst = num_inst
        self._extra_stop_points = extra_stop_points

        # Lift and process the block to get successors, retry if necessary
        finished = False
        while not finished:
            self._lift_irsb()
            self._probe_access()
            self._store_successor_artifacts(successors)
            finished = self._process_irsb()

        self._process_successor_exits(successors)
        successors.processed = True
        return None

    def _lift_irsb(self):
        irsb = self.state.scratch.irsb
        if irsb is None:
            irsb = self.lift_pcode(
                addr=self._addr,
                state=self.state,
                insn_bytes=self._insn_bytes,
                thumb=self._thumb,
                size=self._size,
                num_inst=self._num_inst,
                extra_stop_points=self._extra_stop_points,
            )
        if irsb.size == 0:
            if irsb.jumpkind == "Ijk_NoDecode" and not self.state.project.is_hooked(irsb.addr):
                raise errors.SimIRSBNoDecodeError(
                    f"IR decoding error at {self._addr:#x}. You can hook this instruction with "
                    "a python replacement using project.hook"
                    f"({self._addr:#x}, your_function, length=length_of_instruction)."
                )
            raise errors.SimIRSBError("Empty IRSB passed to HeavyPcodeMixin.")
        self.state.scratch.irsb = irsb

    def _store_successor_artifacts(self, successors: SimSuccessors) -> None:
        """
        Update successors.artifacts with IRSB details.
        """
        irsb = self.state.scratch.irsb
        successors.artifacts["irsb"] = irsb
        successors.artifacts["irsb_size"] = irsb.size
        successors.artifacts["irsb_direct_next"] = irsb.direct_next
        successors.artifacts["irsb_default_jumpkind"] = irsb.jumpkind
        successors.artifacts["insn_addrs"] = []

    def _probe_access(self) -> None:
        """
        Check permissions, are we allowed to execute here? Do we care?
        """
        if o.STRICT_PAGE_ACCESS not in self.state.options:
            return
        try:
            perms = self.state.memory.permissions(self._addr)
        except errors.SimMemoryError:
            raise errors.SimSegfaultError(self._addr, "exec-miss") from errors.SimMemoryError
        else:
            if not perms.symbolic:
                perms = self.state.solver.eval(perms)
                if not perms & 4 and o.ENABLE_NX in self.state.options:
                    raise errors.SimSegfaultError(self._addr, "non-executable")

    def _process_irsb(self) -> bool:
        """
        Execute the IRSB. Returns True if successfully processed.
        """
        try:
            self.handle_pcode_block(self.state.scratch.irsb)
            return True
        except errors.SimReliftException as e:
            self.state = e.state
            if self._insn_bytes is not None:
                raise errors.SimEngineError("You cannot pass self-modifying code as insn_bytes!!!") from e
            new_ip = self.state.scratch.ins_addr
            if self._size is not None:
                self._size -= new_ip - self._addr
            if self._num_inst is not None:
                self._num_inst -= self.state.scratch.num_insns
            self._addr = new_ip

            # clear the stage before creating the new IRSB
            self.state.scratch.dirty_addrs.clear()
            self.state.scratch.irsb = None
        except errors.SimError as ex:
            ex.record_state(self.state)
            raise
        # FIXME:
        # except VEXEarlyExit:
        #     break
        return False

    def _process_successor_exits(self, successors: SimSuccessors) -> None:
        """
        Do return emulation and call-less stuff.
        """
        for exit_state in list(successors.all_successors):
            exit_jumpkind = exit_state.history.jumpkind
            if exit_jumpkind is None:
                exit_jumpkind = ""

            if o.CALLLESS in self.state.options and exit_jumpkind == "Ijk_Call":
                # get the default calling convention for the architecture and retrieve the return value offset
                if exit_state.arch.name in DEFAULT_CC:
                    cc = default_cc(
                        exit_state.arch.name,
                        platform=self.project.simos.name if self.project.simos is not None else None,
                    )
                    ret_reg = cc.RETURN_VAL
                    if isinstance(ret_reg, SimRegArg):
                        ret_offset = exit_state.arch.registers[ret_reg.reg_name][0]
                        exit_state.registers.store(
                            ret_offset,
                            exit_state.solver.Unconstrained("fake_ret_value", exit_state.arch.bits),
                        )
                    else:
                        if once("return_val_is_not_reg"):
                            l.warning(
                                "The return value of the default calling convention for architecture %s is not "
                                "stored in a register. We cannot set the fake return value in Call-less mode. "
                                "Please report to GitHub.",
                                exit_state.arch.name,
                            )
                else:
                    if once("missing_default_cc"):
                        l.warning(
                            "Default calling convention is not set for architecture %s. We cannot set the fake "
                            "return value in Call-less mode.",
                            exit_state.arch.name,
                        )
                exit_state.scratch.target = claripy.BVV(
                    successors.addr + self.state.scratch.irsb.size, exit_state.arch.bits
                )
                exit_state.history.jumpkind = "Ijk_Ret"
                exit_state.regs.ip = exit_state.scratch.target
                if exit_state.arch.call_pushes_ret:
                    exit_state.regs.sp = exit_state.regs.sp + exit_state.arch.bytes

            elif o.DO_RET_EMULATION in exit_state.options and (
                exit_jumpkind == "Ijk_Call" or exit_jumpkind.startswith("Ijk_Sys")
            ):
                l.debug("%s adding postcall exit.", self)

                ret_state = exit_state.copy()
                guard = claripy.true() if o.TRUE_RET_EMULATION_GUARD in self.state.options else claripy.false()
                ret_target = claripy.BVV(successors.addr + self.state.scratch.irsb.size, ret_state.arch.bits)
                if ret_state.arch.call_pushes_ret and not exit_jumpkind.startswith("Ijk_Sys"):
                    ret_state.regs.sp = ret_state.regs.sp + ret_state.arch.bytes
                successors.add_successor(
                    ret_state,
                    ret_target,
                    guard,
                    "Ijk_FakeRet",
                    exit_stmt_idx=DEFAULT_STATEMENT,
                    exit_ins_addr=self.state.scratch.ins_addr,
                )



from __future__ import annotations

import angr


class strxfrm(angr.SimProcedure):
    """Transform string for locale-aware comparison."""

    # pylint:disable=arguments-differ

    def run(self, dest, src, n):
        strlen = angr.SIM_PROCEDURES["libc"]["strlen"]
        strncpy = angr.SIM_PROCEDURES["libc"]["strncpy"]
        src_len = self.inline_call(strlen, src)

        # Defaults to strcpy, if the current locale is "C" or "POSIX".
        self.inline_call(strncpy, dest, src, n, src_len=src_len.ret_expr)

        return src_len.ret_expr
